/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * Assorted debugging-related code.
 *
 * Also error code support.
 */

typedef void (*report_func_t)(char *msg, ...);

typedef struct _error_code_t
{
    int code;
    report_func_t report_func;
    char *description;
} error_code_t;

static error_code_t fm_errors[] = {

    {FM_IOSTAT_EOF,	(report_func_t) NULL,
	 "Channel closed"},

    {FM_IOSTAT_OK,	(report_func_t) NULL,
	 "No error"},

    {FM_IOSTAT_ERROR,	yoke_fatal,
	 "Unspecified error"},

    {FM_IOSTAT_INPORT_UNCONNECTED, yoke_fatal,
	 "Access to an unconnected inport"},

    {FM_IOSTAT_OUTPORT_UNCONNECTED, yoke_fatal,
	 "Access to an unconnected outport"},

    {FM_IOSTAT_INPORT_ALREADY_CONNECTED, yoke_fatal,
	 "Attempt to connect an already connected inport"},

    {FM_IOSTAT_OUTPORT_ALREADY_CONNECTED, yoke_fatal,
	 "Attempt to connect an already connected outport"},

    {FM_IOSTAT_CLOSE_TWICE, yoke_warning,
	 "Closing an outport twice"},
};

#define n_fm_errors (sizeof(fm_errors) / sizeof(error_code_t))

static void set_iostat_guts(f_integer *iostat, int have_iostat, int error,
			    report_func_t *report_func, char **descr);

/*
 * _fm_set_iostat()
 *
 * Set the IOSTAT variable. If none was passed from the Fortran M
 * code (have_iostat == FALSE), print an error message. If this error
 * is marked as a fatal error, abort the Fortran M program if no IOSTAT was
 * passed.
 */
void _fm_set_iostat(f_integer *iostat, int have_iostat, int error)
{
    int i;
    report_func_t report_func;
    char *descr;

    set_iostat_guts(iostat, have_iostat, error, &report_func, &descr);
    if (report_func != (report_func_t) NULL)
	(*report_func)("Runtime error %d: %s\n", error, descr);
}

#ifdef DEBUG

void _fm_set_iostat_debug(f_integer *iostat, int have_iostat, int error,
			  char *file, int line)
{
    int i;
    report_func_t report_func;
    char *descr;

    set_iostat_guts(iostat, have_iostat, error, &report_func, &descr);
    if (report_func != (report_func_t) NULL)
	(*report_func)("Runtime error %d: %s (at %s:%d)\n",
		       error, descr, file, line);
}

#endif

static void set_iostat_guts(f_integer *iostat, int have_iostat, int error,
			    report_func_t *report_func, char **descr)
{
    int i;

    *iostat = (f_integer) error;

    *descr = (char *) NULL;
    for (i = 0; i < n_fm_errors; i++)
	if (fm_errors[i].code == error)
	{
	    *report_func = fm_errors[i].report_func;
	    *descr = fm_errors[i].description;
	    break;
	}
    if (*descr == (char *) NULL)
    {
	*report_func = yoke_fatal;
	*descr = "Unknown error";
    }
}

static fm_pausepoint_t lookup_pausepoint(int id)
{
    fm_pausepoint_t ppoint;

    for (ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next)
    {
	if (ppoint->id == id)
	    return ppoint;
    }
    return (fm_pausepoint_t) NULL;
}


void _fm_sizeof_pausepoints(conn_t dest, int *size, int *count)
{
    fm_pausepoint_t ppoint;
    int n;

    for (n = 0, ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next, n++)
	;

    /* For the number of pausepoints */
    *size = yoke_sizeof_int(dest, n * 3 + 1);
    *count = n * 3 + 1;
}

void _fm_put_pausepoints(yoke_buffer_t buf)
{
    fm_pausepoint_t ppoint;
    int n;

    for (n = 0, ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next, n++)
	;

    yoke_put_int(buf, &n, 1);

    for (ppoint = _fm_pausepoints; ppoint != (fm_pausepoint_t) NULL;
	 ppoint = ppoint->next)
    {
	yoke_put_int(buf, &ppoint->id, 1);
	yoke_put_int(buf, &ppoint->pause_me, 1);
	yoke_put_int(buf, &ppoint->sleep_me, 1);
    }
}

void _fm_get_pausepoints(yoke_buffer_t buf)
{
    int n;
    int i;
    fm_pausepoint_t ppoint;

    yoke_get_int(buf, &n, 1);

    for (i = 0; i < n; i++)
    {
	ppoint = (fm_pausepoint_t) malloc(sizeof( struct _fm_pausepoint_t));
	MallocCheck(_fm_add_pausepoint, ppoint, fm_pausepoint_t);

	yoke_get_int(buf, &ppoint->id, 1);
	yoke_get_int(buf, &ppoint->pause_me, 1);
	yoke_get_int(buf, &ppoint->sleep_me, 1);

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
}

void _fm_add_pausepoint(int id)
{
    fm_pausepoint_t ppoint;

    ppoint = lookup_pausepoint(id);

    if (ppoint == (fm_pausepoint_t) NULL)
    {
	ppoint = (fm_pausepoint_t) malloc(sizeof(struct _fm_pausepoint_t));
	MallocCheck(_fm_add_pausepoint, ppoint, fm_pausepoint_t);

	ppoint->id = id;
	ppoint->pause_me = FALSE;
	ppoint->sleep_me = 0;

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
    
    ppoint->pause_me = TRUE;
    
}

void _fm_add_sleeppoint(int id, int secs)
{
    fm_pausepoint_t ppoint;

    ppoint = lookup_pausepoint(id);

    if (ppoint == (fm_pausepoint_t) NULL)
    {
	ppoint = (fm_pausepoint_t) malloc(sizeof(struct _fm_pausepoint_t));
	MallocCheck(_fm_add_pausepoint, ppoint, fm_pausepoint_t);

	ppoint->id = id;
	ppoint->pause_me = FALSE;
	ppoint->sleep_me = 0;

	ppoint->next = _fm_pausepoints;
	_fm_pausepoints = ppoint;
    }
    
    ppoint->sleep_me = secs;
    
}

void FM_PAUSE(f_integer *f_id, f_character f_msg CHAR_LEN_DECL(f_msg_len))
{
    char msg[2048];
    int id;
    LOCAL_CHAR_LEN_DECL(f_msg_len);

    CONVERT_CHAR_FTN_TO_C_INPLACE(f_msg, f_msg_len, msg);

    id = (int) *f_id;

    _fm_pause(id, msg);
}

void _fm_pause(int id, char *msg)
{
    fm_pausepoint_t ppoint;
    int done_pausing = 0;

    ppoint = lookup_pausepoint(id);

    if (_fm_show_pausepoints)
    {
	yoke_printf("Pausepoint %d: %s\n", id, msg);
    }
    
    if (ppoint != (fm_pausepoint_t) NULL)
    {
	if (ppoint->pause_me)
	{
	    yoke_printf("Pausing at pausepoint %d: %s\n", id, msg);
	    while (!done_pausing)
	    {
		pause();
	    }
	    yoke_printf("Continuing from pausepoint %d\n", id);
	}
	else if (ppoint->sleep_me > 0)
	{
	    yoke_printf("Sleeping %d at pausepoint %d: %s\n", ppoint->sleep_me, id, msg);
	    sleep(ppoint->sleep_me);
	    yoke_printf("Continuing from pausepoint %d\n", id);
	}
    }
}

char *_fm_mtype_name(int type)
{
    int chan;
    static char str[5][1024];
    static int which = 0;
    static char *buf;
    static int maxtype;
    static struct
    {
	char *name;
	int type;
    } msg_types[] = {
#include "msgtype_names.h"
	(char *) NULL, 0,
    };
    static char **typenames = NULL;

    buf = str[which];

    which = ((which + 1) % 5);

    if (typenames == NULL)
    {
	int i;

	maxtype = -1;
	for (i = 0; msg_types[i].name != (char *) NULL; i++)
	{
	    if (msg_types[i].type > maxtype)
		maxtype = msg_types[i].type;
	}
	typenames = (char **) malloc(sizeof(char *) * (maxtype+1));
	MallocCheck(_fm_mtype_name, typenames, char **);

	ZeroOutMemory(typenames, sizeof(char *) * (maxtype + 1));
	for (i = 0; msg_types[i].name != (char *) NULL; i++)
	{
	    typenames[msg_types[i].type] = msg_types[i].name;
	}
    }
    
    chan = FM_CHANNEL_OF_TYPE(type);
    if (FM_IS_SEND(type))
	sprintf(buf, "SEND(%d)", chan);
    else if (FM_IS_CLOSE(type))
	sprintf(buf, "CLOSE(%d)", chan);
    else if (FM_IS_FWD_BUF(type))
	sprintf(buf, "FWD_BUF(%d)", chan);
    else if (FM_IS_FWD_LAST_BUF(type))
	sprintf(buf, "FWD_LAST_BUF(%d)", chan);
    else
    {
	if (chan < 0 || chan > maxtype)
	    sprintf(buf, "invalid(%d)", chan);
	else
	    strcpy(buf, typenames[chan]);
    }
    return buf;
}


char *_fm_outport_as_string(fm_outport_t outport)
{
    static char buf[3][1024];
    static int which = 0;
    char *rc, *name, *colon;

    rc = buf[which];
    which = (which + 1) % 3;

#ifdef DEBUG
    name = outport->name;
    colon = ":";
#else
    name = "";
    colon = "";
#endif


    sprintf(rc, "<%x:%s%s%d:%s:%x:%s:%d:%c>",
	    (unsigned int) outport,
	    name,
	    colon,
	    outport->channel_id,
	    yoke_conn_as_string(outport->channel_conn),
	    (unsigned int) outport->local_inport,
	    yoke_conn_as_string(outport->future_home),
	    outport->inport_index,
	    outport->closed ? 'T' : 'F');
    return rc;
}

char *_fm_inport_as_string(fm_inport_t inport)
{
    static char buf[3][1024];
    static int which = 0;
    static char outbuf[1024], obuf[1024];
    char *rc, *name, *colon;
    int i;
    fm_outport_reference_t *ref;

    rc = buf[which];
    which = (which + 1) % 3;

#ifdef DEBUG
    name = inport->name;
    colon = ":";
#else
    name = "";
    colon = "";
#endif

    outbuf[0] = '\0';
    for (i = 0; i < inport->n_outports; i++)
    {
	ref = &(inport->outports[i]);
	sprintf(obuf, "%x=%s",
		(unsigned int) ref->local_outport,
		ref->remote_outport ?
		     yoke_conn_as_string(ref->remote_outport) :
		     "<none>");
	strcat(outbuf, obuf);
	if (i < inport->n_outports-1)
	    strcat(outbuf, ",");
    }

    sprintf(rc, "<%x:%s%s%d:%s:%d:%s>",
	    (unsigned int) inport,
	    name,
	    colon,
	    inport->channel_id,
	    yoke_conn_as_string(inport->channel_conn),
	    inport->n_outports,
	    outbuf);
    return rc;
}
	    

