/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * FM Manager.
 *
 * This module encapsulates the actions of the manager processes.
 * Each manager keeps a list and hashtable of all children executing
 * under it.
 *
 * The actions of the manager include
 *
 *    Process creation on behalf of Fortran M processes, including
 *        the initial assignment of channel ids.
 *    Monitoring of the death of child processes
 *    Signalling of child processes for port migration
 */

typedef struct _child_t
{
    int pid;
    conn_t conn;
    bool_t exited;
    bool_t exited_normally;
    struct _child_t *prev, *next;
    struct _child_t *h_prev, *h_next;
} *child_t;

/*
 * The list of all children.
 */
static child_t children;

/*
 * The hashtable of all children
 */

#define CHILDREN_HASHTABLE_SIZE 1009
static struct _child_t children_hashtable[CHILDREN_HASHTABLE_SIZE];

static bool_t cleaning_children;

/*
 * Local functions.
 */
static child_t alloc_child();
static child_t lookup_child(int pid);
static void hashtable_add_child(child_t c);
static void create_process(yoke_buffer_t buf, conn_t sender);
static void signal_child(yoke_buffer_t buf, conn_t sender);
static void process_wait_for_start();
static void child_exiting(yoke_buffer_t buf, conn_t sender);
static void send_manager_list(yoke_buffer_t buf, conn_t sender);
static void do_outport_migration(yoke_buffer_t buf, conn_t sender);
static void do_inport_migration(yoke_buffer_t buf, conn_t sender);

static void child_death_func(int pid, int signal, int retcode);
static bool_t clean_up_msg_loop();
static void mark_child_exited(int child_pid);

/*
 * _fm_manager_init()
 *
 * Initialize the manager data structures.
 */
void _fm_manager_init()
{
    int i;
    child_t ent;

    _fm_process_name = (char *) malloc(strlen("manager:") + 10);
    MallocCheck(_fm_manager_init(), _fm_process_name, char *);
    sprintf(_fm_process_name, "mananger:%d", _fm_manager_id);
    _fm_parent_process_name = (char *) malloc(strlen("toplevel") + 1);
    MallocCheck(_fm_manager_init(), _fm_parent_process_name, char *);
    strcpy(_fm_parent_process_name, "toplevel");

    /*
     * The children list is a circular linked list of child_t structures
     * with a dummy header.
     */
    children = alloc_child(-1, NULL);
    children->next = children->prev = children;

    /*
     * Channel ids start at zero.
     */
    _fm_next_channel_id = 0;

    /*
     * The children hash table contains circular doubly linked lists
     * of buckets with dummy headers kept in children_hashtable.
     */
    for (i = 0, ent = children_hashtable; i < CHILDREN_HASHTABLE_SIZE; i++, ent++)
    {
	ent->pid = -1;
	ent->conn = NULL;
	ent->prev = ent->next = NULL;
	ent->h_prev = ent->h_next = ent;
    }

    cleaning_children = FALSE;

    yoke_register_child_death_func(child_death_func);
}


/*
 * _fm_manager_run()
 *
 * Manager main loop. We spend most of our time waiting in a
 * yoke_recv_transient.
 */
void _fm_manager_run()
{
    yoke_buffer_t buf;
    int type, child_pid;
    conn_t sender;

#ifdef DEBUG
    if (FMDebug(2))
	yoke_printf("Manager running self=%s\n",
		    yoke_conn_as_string(yoke_self())); 
#endif

    while (1)
    {
#ifdef DEBUG
	if (FMDebug(3))
	    yoke_printf("Manager waiting for message\n");
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (FMDebug(3))
	    yoke_printf("Manager got message type %d %s\n",
			type, _fm_mtype_name(type));
#endif

	switch (type)
	{
	case FM_CREATE_PROCESS:
	    create_process(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_SIGNAL_CHILD:
	    signal_child(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_MTYPE_MANAGER_LIST_REQ:
	    send_manager_list(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_MTYPE_INPORT_MGR_MIGRATE:
	    do_inport_migration(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;
	    
	case FM_MTYPE_OUTPORT_MGR_MIGRATE:
	    do_outport_migration(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_MTYPE_CHILD_EXITING:
	    child_exiting(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_MTYPE_SHUTDOWN_REQUEST:
	    yoke_get_int(buf, &child_pid, 1);
#ifdef DEBUG
	    if (FMDebug(2))
		yoke_printf("Manager %d got shutdown request from child %d\n",
			    _fm_manager_id, child_pid);
#endif
	    yoke_free_buffer(buf);
	    buf = yoke_allocate_buffer(sender, 0, 0);

	    mark_child_exited(child_pid);

	    yoke_send(sender, buf, FM_MTYPE_SHUTDOWN_REQUEST_ACK);
	    yoke_release_transient(sender);

	    if (_fm_manager_id == 0)
	    {
		_fm_handle_shutdown_request();
	    }
	    else
	    {
		yoke_buffer_t nbuf;

		nbuf = yoke_allocate_buffer(_fm_manager_list[0], 0, 0);
		yoke_establish(_fm_manager_list[0]);
		yoke_send(_fm_manager_list[0], nbuf, FM_MTYPE_DO_SHUTDOWN);
		yoke_release(_fm_manager_list[0]);
	    }
	    break;

	case FM_MTYPE_ABEND:
	    Assert(_fm_manager_id == 0);
	    _fm_handle_abend(buf);
	    break;
	    
	case FM_MTYPE_DO_SHUTDOWN:
	    Assert(_fm_manager_id == 0);
	    _fm_handle_shutdown_request();
	    break;

	case FM_MTYPE_MGR_SHUTDOWN_1:
#ifdef DEBUG
	    if (FMDebug(2))
		yoke_printf("Manager %d got MGR_SHUTDOWN_1\n", _fm_manager_id);
#endif
	    yoke_free_buffer(buf);
	    _fm_shutdown_one_manager(sender);
	    break;

	default:
	    yoke_warning("Manager got invalid message type %d\n", type);
	    break;
	}

    }
}


/*
 * create_process()
 *
 * Create a process for a new Fortran M process. We are passed the
 * number of channels the new process requires, an identifier for the
 * procedure which should be invoked, the new vp_id and vc_offset for the
 * new process, the conn_t for the creating process, and the slot
 * this process is assigned in the creator's process block.
 *
 * We allocate the number of channels requested. This merely involves
 * bumping the value of _fm_next_channel_id by the number of channels
 * requested. Then fork.
 *
 * In the child:
 *     initialize the process state,
 *     close off any lingering fds in the manager list. This has to get
 *        cleaned up
 *     create the actual channel data structures
 *     wait for the invocation message from the creating process.
 *
 * In the parent:
 *     send the conn_t and channel ids for the new process back to the creator
 *     add the new child to the manager child list.
 *
 */
static void create_process(yoke_buffer_t buf, conn_t sender)
{
    int n_ports, vc_base, procedure_ident;
    conn_t him, creator;
    int child_pid, channel_start, vp_id, slot;
    yoke_buffer_t rbuf;
    int type;
    int remote_pblock;
    int *port_sizes;
    char *subprogram_name, *parent_name;
    int len, block_number;

    yoke_get_int(buf, &n_ports, 1);
    if (n_ports > 0)
    {
	port_sizes = (int *) malloc(sizeof(int) * n_ports);
	MallocCheck(create_process(), port_sizes, int *);
    }
    else
	port_sizes = (int *) NULL;
    yoke_get_int(buf, port_sizes, n_ports);

    yoke_get_int(buf, &len, 1);
    subprogram_name = (char *) malloc(len);
    MallocCheck(create_process(), subprogram_name, char *);
    yoke_get_char(buf, subprogram_name, len);
    
    yoke_get_int(buf, &len, 1);
    parent_name = (char *) malloc(len);
    MallocCheck(create_process(), parent_name, char *);
    yoke_get_char(buf, parent_name, len);
    yoke_get_int(buf, &block_number, 1);
    
    yoke_get_int(buf, &procedure_ident, 1);
    yoke_get_int(buf, &vp_id, 1);
    yoke_get_int(buf, &vc_base, 1);
    yoke_get_conn(buf, &creator);
    yoke_get_int(buf, &remote_pblock, 1);
    yoke_get_int(buf, &slot, 1);
#ifdef DEBUG
    if (ProcessDebug(3))
    {
	yoke_printf("Got request to create %d ports procedure_ident=%d pid=%d vc_offset=%d slot=%d parent=<%s> subprogram=<%s> block=%d\n",
		    n_ports, procedure_ident, vp_id, vc_base, slot,
		    parent_name, subprogram_name, block_number);
    }
#endif    

    channel_start = _fm_next_channel_id;
    _fm_next_channel_id += n_ports;

    him = yoke_fork(&child_pid);

    if (child_pid == 0)
    {
	char *process;
	int pname_len;
	
	/*
	 * We're the new child. Initialize the process data structures,
	 * setting the virtual processor id and virtual computer base
	 * to those passed in the message. We cannot set the VC size until
	 * we get into the user's process code, where a call to
	 * FM_SET_VC_SIZE() is made using the information in that
	 * process' PROCESSORS declaration.
	 */

	/*
	 * The new name is
	 *
	 *  <parent>:<subprogram>/<block>/<slot>:<process>
	 *
	 * We'll allocate lotsa room for the numbers :-)
	 */

	process = _fm_lookup_process(procedure_ident);

	pname_len = strlen(process) + strlen(subprogram_name);
	pname_len += strlen(parent_name);
	pname_len += 5 /* for the separators and null */;
	pname_len += 2 * 8 /* for the numbers */;

	if (_fm_process_name != (char *) NULL)
	    free(_fm_process_name);

	_fm_process_name = (char *) malloc(pname_len);
	MallocCheck(create_process(), _fm_process_name, char *);

	sprintf(_fm_process_name,
		"%s:%s/%d/%d:%s",
		parent_name,
		subprogram_name,
		block_number,
		slot + 1,
		process);

	_fm_parent_process_name = parent_name;

	free(subprogram_name);

	_fm_process_init(vp_id, vc_base);
	_fm_process_setup_channels(channel_start, n_ports, port_sizes);
	if (port_sizes != (int *) NULL)
	    free(port_sizes);
	_fm_process_slot = slot;
	_fm_process_remote_pblock = remote_pblock;
#ifdef DEBUG
	if (ProcessDebug(2))
	    yoke_printf("Waiting for process start id=%d func=%s slot=%d \n",
			procedure_ident,
			_fm_lookup_remote_start_name(procedure_ident),
			_fm_process_slot);
#endif	
	process_wait_for_start(procedure_ident);
	exit(0);
    }

    free(subprogram_name);

    if (port_sizes != (int *) NULL)
	free(port_sizes);


#ifdef DEBUG
    if (FMDebug(3))
    {
	yoke_printf("Fork returned new conn %s\n", yoke_conn_as_string(him));
    }
#endif

    /*
     * Send the new child his manager's conn_t and creator.
     */

    rbuf = yoke_allocate_buffer(him,
				yoke_sizeof_conn(him, yoke_self()) +
				yoke_sizeof_conn(him, creator), 2);
    yoke_put_conn(rbuf, yoke_self());
    yoke_put_conn(rbuf, creator);
    yoke_establish(him);
    yoke_send(him, rbuf, FM_MTYPE_YOUR_MANAGER);
    rbuf = yoke_recv(him, &type);
    Assert(type == FM_MTYPE_YOUR_MANAGER_ACK);
    yoke_release(him);
    yoke_free_buffer(rbuf);

    /*
     * Reply to the request, returning the connection for the new child
     * and the start of his channels.
     */
    rbuf = yoke_allocate_buffer(sender, yoke_sizeof_int(sender, 1) +
				 yoke_sizeof_conn(sender, him),
				2);
    yoke_put_conn(rbuf, him);
    yoke_put_int(rbuf, &channel_start, 1);

    yoke_send(sender, rbuf, FM_CREATE_PROCESS_RETURN);

    _fm_manager_add_child(child_pid, him);
}


/*
 * _fm_manager_clean_up_children()
 *
 * Walk the list of children looking for any that
 * haven't died yet. If there are any, commit
 * infanticide.
 *
 * Return true if we have received a FM_MTYPE_MGR_SHUTDOWN_2
 * message.
 */
bool_t _fm_manager_clean_up_children()
{
    child_t child;
    bool_t try_again;
    bool_t got_shutdown_2 = FALSE;
    
#ifdef DEBUG
    if (FMDebug(3))
	yoke_printf("_fm_manager_clean_up_children(): Manager %d cleaning up\n",
		    _fm_manager_id);
#endif

    cleaning_children = TRUE;

    try_again = FALSE;
    for (child = children->next; child != children; child = child->next)
    {
	if (!child->exited)
	{
	    try_again = TRUE;
	}
    }
    
    if (try_again)
    {
	double now, cutoff;

	now = yoke_wallclock();
	cutoff = now + 3.0;

	try_again = TRUE;
	while (try_again && ((now = yoke_wallclock()) < cutoff))
	{
	    if (clean_up_msg_loop())
		got_shutdown_2 = TRUE;
	    try_again = FALSE;
	    for (child = children->next; child != children; child = child->next)
	    {
		if (!child->exited)
		{
		    try_again = TRUE;
		    break;
		}
	    }
	    yoke_usleep(100000);
	}

	for (child = children->next; child != children; child = child->next)
	  {
#ifdef DEBUG
	    if (ProcessDebug(3))
		yoke_printf("Looking at child %x pid=%d conn=%x exited=%d exited_normally=%d\n",
			    child,
			    child->pid,
			    child->conn,
			    child->exited,
			    child->exited_normally);
#endif
	    if (!child->exited)
	    {
		/*
		 * SGI doesn't seem to pick up the SIGCHLD
		 * for a normally terminating child... for
		 * now we'll assume that the terminate worked
		 * ok.
		 *
		 * Looks like the cray does this too.
		 */
#if defined(sgi) || defined(CRAY)
		if (child->exited_normally)
		    continue;
#endif
		yoke_warning("Process %d did not exit\n",
			     child->pid);

		yoke_kill(child->pid);
	    }
	}
    }
    return got_shutdown_2;
}




/*
 * clean_up_msg_loop()
 *
 * Receive any child_exiting messages. Also note if we
 * get a FM_MTYPE_MGR_SHUTDOWN_2 message and return true
 * if we did.
 */
static bool_t clean_up_msg_loop()
{
    yoke_buffer_t buf;
    conn_t sender;
    int type, child_pid;
    bool_t got_shutdown_2 = FALSE;

    while ((buf = yoke_recv_transient(&sender, &type, TRUE)) != (yoke_buffer_t) NULL)
    {
	switch (type)
	{
	case FM_MTYPE_CHILD_EXITING:
	    child_exiting(buf, sender);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    break;

	case FM_MTYPE_MGR_SHUTDOWN_2:
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	    got_shutdown_2 = TRUE;
	    break;
	    
	case FM_MTYPE_SHUTDOWN_REQUEST:
	    yoke_get_int(buf, &child_pid, 1);
#ifdef DEBUG
	    if (FMDebug(2))
		yoke_printf("Manager %d got shutdown request from child %d\n",
			    _fm_manager_id, child_pid);
#endif
	    yoke_free_buffer(buf);
	    buf = yoke_allocate_buffer(sender, 0, 0);

	    mark_child_exited(child_pid);

	    yoke_send(sender, buf, FM_MTYPE_SHUTDOWN_REQUEST_ACK);
	    yoke_release_transient(sender);

	    break;

	default:
#ifdef DEBUG
	    if (ProcessDebug(2))
		yoke_printf("Tossing message type %s %d\n",
			    _fm_mtype_name(type), type);
#endif
	    break;
	}
    }
    return got_shutdown_2;
}


/*
 * mark_child_exited()
 *
 * Given a pid, mark that child as exiting normally.
 */
static void mark_child_exited(int child_pid)
{
    child_t child;

    child = lookup_child(child_pid);

    if (child)
    {
#ifdef DEBUG
	if (ProcessDebug(2))
	    yoke_printf("mark_child_exited(): Found child %x conn=%s\n",
			child, yoke_conn_as_string(child->conn));
#endif
	child->exited_normally = TRUE;
    }
    else
    {
	yoke_fatal("mark_child_exited(): Didn't find child record for pid %d\n",
		   child_pid);
    }
}


/*
 * child_exiting()
 *
 * A child that is about to terminate normally sends this to the
 * manager to notify it that the child's death is normal
 */
static void child_exiting(yoke_buffer_t buf, conn_t sender)
{
    int pid;

    yoke_get_int(buf, &pid, 1);

#ifdef DEBUG
    if (ProcessDebug(2))
	yoke_printf("child_exiting(): Child %d exiting\n",
		    pid);
#endif

    mark_child_exited(pid);

    buf = yoke_allocate_buffer(sender, 0, 0);
    yoke_send(sender, buf, FM_MTYPE_CHILD_EXITING_ACK);
}

static void child_death_func(int pid, int signal, int retcode)
{
    child_t child;
    
#ifdef DEBUG
    if (ProcessDebug(2))
	yoke_printf("child_death_func(): pid=%d signal=%d retcode=%d\n",
		    pid, signal, retcode);
#endif

    child = lookup_child(pid);

    if (child)
    {
	child->exited = TRUE;
	
	if (child->exited_normally)
	{
#ifdef DEBUG
	    if (ProcessDebug(2))
		yoke_printf("Child %d exited normally\n", pid);
#endif
	}
	else if (!cleaning_children)
	{
	    yoke_buffer_t buf;
	    int len;
	    static char msg_buf[MAX_ABEND_MSG_SIZE];
	    
#ifdef DEBUG
	    if (ProcessDebug(2))
		yoke_warning("Child %d exited abnormally signal=%d retcode=%d\n", pid, signal, retcode);
#endif

	    if (_fm_manager_id == 0)
	    {
		yoke_printf("Manager 0: Process %d exited abnormally: signal=%d retcode=%d\n", pid, signal, retcode);
		_fm_handle_shutdown_request();
	    }
	    else
	    {
		sprintf(msg_buf, "Process %d exited abnormally", pid);
		len = strlen(msg_buf) + 1;
		buf = yoke_allocate_buffer(_fm_manager_list[0],
					   yoke_sizeof_int(_fm_manager_list[0], 1) +
					   yoke_sizeof_char(_fm_manager_list[0], len),
					   2);
		yoke_put_int(buf, &len, 1);
		yoke_put_char(buf, msg_buf, len);
		yoke_establish(_fm_manager_list[0]);
		yoke_send(_fm_manager_list[0], buf, FM_MTYPE_ABEND);
		yoke_release(_fm_manager_list[0]);
	    }
	}
    }
    else
    {
	yoke_warning("child_exiting(): Didn't find child record for pid %d\n",
		   pid);
    }
} 


/*
 * send_manager_list()
 *
 * Send the list of managers to our child. This is in response
 * to a _fm_get_manager() call when we don't have the manager list.
 */
static void send_manager_list(yoke_buffer_t buf, conn_t sender)
{
    yoke_buffer_t rbuf;
    int i, size;

    size = yoke_sizeof_int(sender, 1);

    for (i = 0; i < _fm_manager_count; i++)
	size += yoke_sizeof_conn(sender, _fm_manager_list[i]);

    rbuf = yoke_allocate_buffer(sender, size, _fm_manager_count + 1);
    yoke_put_int(rbuf, &_fm_manager_count, 1);
    for (i = 0; i < _fm_manager_count; i++)
    {
	yoke_put_conn(rbuf, _fm_manager_list[i]);
    }
    yoke_send(sender, rbuf, FM_MTYPE_MANAGER_LIST_REQ_REPLY);
}

/*
 * signal_child()
 *
 * Send the FM interrupt signal to the child specified in the
 * message.
 */
static void signal_child(yoke_buffer_t buf, conn_t sender)
{
    yoke_buffer_t nbuf;
    conn_t child_conn;
    char *child_host;
    int child_pid;
    child_t child;
    int failed;

    yoke_get_conn(buf, &child_conn);
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("Manager notifying child %s\n",
		    yoke_conn_as_string(child_conn));
#endif

    yoke_get_conn_id(child_conn, &child_host, &child_pid);

    child = lookup_child(child_pid);
    if (child == NULL || child->exited || child->exited_normally)
    {
#ifdef DEBUG
	if (FMDebug(1))
	    yoke_printf("Manager: signal_child() cannot find child %s\n",
		     yoke_conn_as_string(child_conn));
#endif
    }
    else
    {
	yoke_notify(child_conn);
    }
}


/*
 * process_wait_for_start()
 *
 * This is invoked in new child processes. We loop on recv_transient(), 
 * enqueueing any messages that come in. When we get a FM_START_PROCESS
 * message, we look up the address of the remote start function and
 * invoke it with the message associated with FM_START_PROCESS (that will
 * contain the arguments for the process).
 */
static void process_wait_for_start(int procedure_ident)
{
    yoke_buffer_t buf;
    conn_t sender;
    int type;
    fm_remote_start_func_t func;
    int done;

    /*
     * Receive Loop
     *
     * The matching EndSafeRegion invocation is in
     * _fm_handle_incoming_migrations(), invoked from the compiler-
     * generated remote start routine.
     */
    StartSafeRegion;

    done = 0;
    while (!done)
    {
	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (FMDebug(3))
	    yoke_printf("process_wait_for_start(): got message type %d %s \n",
			type, _fm_mtype_name(type));
#endif

	/*
	 * We also enqueue SEND and CLOSE messages along with the
	 * migration messages in order to ensure proper ordering
	 * of messages.
	 */
	if (FM_IS_SEND(type))
	{
	    _fm_enqueue_migration_request(buf, type);
	    yoke_release_transient(sender);
	}
	else if (FM_IS_CLOSE(type))
	{
	    _fm_enqueue_migration_request(buf, type);
	    yoke_release_transient(sender);
	}
	else
	{
	    switch(type)
	    {
	    case FM_MTYPE_CLOSE_ACK:
		HandleCloseAck(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_INPORT_MIGRATE:
	    {
		conn_t old_conn;
		yoke_buffer_t nbuf;
		
		yoke_get_conn(buf, &old_conn);
		nbuf = yoke_allocate_buffer(old_conn, 0, 0);
#ifdef DEBUG
		if (MigrateDebug(2))
		    yoke_printf("process_wait_for_start(): sending INPORT_MIGRATE_ACK to %s\n",
				yoke_conn_as_string(old_conn));
#endif
		yoke_establish(old_conn);
		yoke_send(old_conn, nbuf, FM_MTYPE_INPORT_MIGRATE_ACK);
		yoke_release(old_conn);
		yoke_reset_buffer(buf);
		_fm_enqueue_migration_request(buf, FM_MTYPE_INPORT_MIGRATE_NO_ACK);
		yoke_release_transient(sender);
		break;
	    }
		
	    case FM_MTYPE_SEND_PORTS:
		EnqueuePortSend(sender, buf);
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
	    {
		conn_t reply_conn;
		yoke_buffer_t nbuf;
		
		yoke_get_conn(buf, &reply_conn);
		nbuf = yoke_allocate_buffer(reply_conn, 0, 0);
#ifdef DEBUG
		if (MigrateDebug(2))
		    yoke_printf("process_wait_for_start(): sending OUTPORT_MIGRATE_ACK to %s\n",
				yoke_conn_as_string(reply_conn));
#endif
		yoke_establish(reply_conn);
		yoke_send(reply_conn, nbuf, FM_MTYPE_OUTPORT_MIGRATE_ACK);
		yoke_release(reply_conn);
		yoke_reset_buffer(buf);
		_fm_enqueue_migration_request(buf, FM_MTYPE_OUTPORT_MIGRATE_NO_ACK);
		yoke_release_transient(sender);

		break;
	    }
		
	    case FM_MTYPE_OUTPORT_PERM_REQ:
	    {
		int channel_id;
		conn_t reply_conn;
		yoke_buffer_t nbuf;

		yoke_get_int(buf, &channel_id, 1);					
		yoke_get_conn(buf, &reply_conn);					

		nbuf = yoke_allocate_buffer(reply_conn, 0, 0);			
#ifdef DEBUG
		if (MigrateDebug(2))
		    yoke_printf("process_wait_for_start(): sending PERM_GRANTED to %s\n",
				yoke_conn_as_string(reply_conn));
#endif
		yoke_establish(reply_conn);
		yoke_send(reply_conn, nbuf, FM_MTYPE_OUTPORT_PERM_GRANTED);	
		yoke_release(reply_conn);
		yoke_release_transient(sender);
		break;
	    }
		
	    case FM_MTYPE_YOUR_MANAGER:
		yoke_get_conn(buf, &_fm_my_manager);
		yoke_get_conn(buf, &_fm_process_creator);
		yoke_free_buffer(buf);
		buf = yoke_allocate_buffer(sender, 0, 0);
		yoke_send(sender, buf, FM_MTYPE_YOUR_MANAGER_ACK);
		yoke_release_transient(sender);
		break;

	    case FM_START_PROCESS:
#ifdef DEBUG
		if (ProcessDebug(2))
		{
		    yoke_printf("Got process start ident=%d name=%s sender: %s\n",
				procedure_ident,
				_fm_lookup_remote_start_name(procedure_ident),
				yoke_conn_as_string(sender));
		}
#endif
		done = 1;
		break;

	    default:
		yoke_fatal("process_wait_for_start() received type %s %d instead of FM_START_PROCESS\n",
			   _fm_mtype_name(type), type);
		break;
	    }
	}
    }
    func = _fm_lookup_remote_start(procedure_ident);

    /*
     * The remote start function frees the buffer before
     * invoking user code
     */
    (*func)(sender, buf);

#ifdef DEBUG
    if (ProcessDebug(2))
	yoke_printf("process_wait_for_start(): procedure %d %s returned with n_close_acks_needed=%d\n",
		    procedure_ident, 
		    _fm_lookup_remote_start_name(procedure_ident),
		    _fm_n_close_acks_needed);
#endif

    while (_fm_n_close_acks_needed > 0)
    {
#ifdef DEBUG
	if (ProcessDebug(2))
	    yoke_printf("process_wait_for_start(): waiting for %d close acks\n",
			_fm_n_close_acks_needed);
#endif
	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (FMDebug(3))
	    yoke_printf("process_wait_for_start(): got message type %d %s while waiting for close acks\n",
			type, _fm_mtype_name(type));
#endif
	/*
	 * Hmm. Do we need to stick around until all of our channels
	 * our closed so we can send close acks so other procs don't
	 * hang? We could migrate the leftover channels to the
	 * manager...
	 */
	if (FM_IS_CLOSE(type))
	{
	    ProcessCloseMessage(type, buf);
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
	}
	else
	{
	    switch (type)
	    {
	    case FM_MTYPE_CLOSE_ACK:
		HandleCloseAck(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_INPORT_MIGRATE:
		HandleInportMigrate(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    case FM_MTYPE_OUTPORT_MIGRATE:
		HandleOutportMigrate(sender, buf);
		yoke_release_transient(sender);
		break;
		
	    default:
		yoke_fatal("During wait for close acks for %s got unknown message type %s %d\n",
			   _fm_lookup_remote_start_name(procedure_ident),
			   _fm_mtype_name(type), type);
	    }
	}
    }
    /*
    EndSafeRegion;
    */
    _fm_process_exit(0);
}


/*
 * do_outport_migration()
 *
 */
static void do_outport_migration(yoke_buffer_t buf, conn_t sender)
{
    yoke_fatal("do_outport_migration(): Not implemented\n");
}


/*
 * do_inport_migration()
 *
 */
static void do_inport_migration(yoke_buffer_t buf, conn_t sender)
{
    conn_t child, new_conn, old_conn, reply_conn;
    int new_id, slot_idx, inport_idx, old_id;
    yoke_buffer_t nbuf;

    yoke_get_conn(buf, &child);
    yoke_get_conn(buf, &reply_conn);
    yoke_get_conn(buf, &old_conn);
    yoke_get_int(buf, &old_id, 1);
    yoke_get_conn(buf, &new_conn);
    yoke_get_int(buf, &new_id, 1);
    yoke_get_int(buf, &slot_idx, 1);
    yoke_get_int(buf, &inport_idx, 1);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("do_inport_migration(): Got request for child=%s new_conn=%s new_id=%d slot_idx=%d inport_idx=%d\n",
		    yoke_conn_as_string(child),
		    yoke_conn_as_string(new_conn),
		    new_id, slot_idx, inport_idx);
#endif

    nbuf = yoke_allocate_buffer(child,
				yoke_sizeof_conn(child, reply_conn) +
				yoke_sizeof_conn(child, old_conn) +
				yoke_sizeof_conn(child, new_conn) +
				yoke_sizeof_int(child, 4), 6);
    yoke_put_conn(nbuf, reply_conn);
    yoke_put_conn(nbuf, old_conn);
    yoke_put_int(nbuf, &old_id, 1);
    yoke_put_conn(nbuf, new_conn);
    yoke_put_int(nbuf, &new_id, 1);
    yoke_put_int(nbuf, &slot_idx, 1);
    yoke_put_int(nbuf, &inport_idx, 1);
    yoke_establish(child);
    yoke_send(child, nbuf, FM_MTYPE_INPORT_MIGRATE);
    yoke_release(child);

    yoke_notify(child);
}


void _fm_manager_add_child(int pid, conn_t conn)
{
    child_t c;

#ifdef DEBUG
    if (ProcessDebug(3))
	yoke_printf("Manager adding child %d\n", pid);
#endif

    c = alloc_child(pid, conn);

    c->pid = pid;
    c->exited = c->exited_normally = FALSE;

    c->prev = children->prev;
    c->next = children;
    children->prev->next = c;
    children->prev = c;

    hashtable_add_child(c);
}


static void hashtable_add_child(child_t c)
{
    int hashval;
    child_t bucket;

    hashval = c->pid % CHILDREN_HASHTABLE_SIZE;
    bucket = &children_hashtable[hashval];

    c->h_next = bucket->h_next;
    c->h_prev = bucket;
    bucket->h_next->h_prev = c;
    bucket->h_next = c;
}

static child_t lookup_child(int pid)
{
    int hashval;
    child_t bucket, c;

    hashval = pid % CHILDREN_HASHTABLE_SIZE;
    bucket = &children_hashtable[hashval];

    for (c = bucket->h_next; c != bucket && c->pid != pid; c = c->h_next)
	;
    if (c == bucket)
	return NULL;
    else
	return c;
}

static child_t alloc_child(int pid, conn_t conn)
{
    child_t rc;

    rc = (child_t) malloc(sizeof(struct _child_t));
    if (rc == (child_t) NULL)
	yoke_fatal("fm_manager:alloc_child(): malloc failed\n");

    rc->pid = pid;
    rc->conn = conn;
    rc->prev = rc->next = NULL;
    return rc;
}

void ___fm_manager_dummies()
{
    (void) lookup_child(0);
}
