/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

static void do_inport_migration_guts(conn_t old_conn, int old_id,
				     conn_t new_conn, int new_id,
				     int inport_index, bool_t send_ack);
static void _fm_do_outport_migration_guts(conn_t reply_conn,
				   conn_t channel_conn,
				   int channel_id,
				   int inport_index,
				   conn_t new_conn, bool_t send_ack);
static void notify_local_outport(fm_inport_t inport,
				 conn_t new_conn, int new_id,
				 fm_outport_t outport);
static void notify_remote_outport(fm_inport_t inport, int inport_index,
				 conn_t new_conn, int new_id,
				 conn_t outport_conn);
static void process_inport_migrate_during_outport_migrate(fm_outport_t outport,
							  int type,
							  yoke_buffer_t buf);
static void send_outport_migration_request(fm_outport_t outport);

static bool_t process_permission_req_checking_for_inport_migrate_ack(yoke_buffer_t, bool_t);

/*
 * _fm_migrate_inport()
 *
 * Migrate this inport.
 *
 * (somewhat badly named since we don't actually do the migration, but
 * rather the setup. It's more obvious in the calling code tho).
 *
 * Notify each outport that this inport is migrating.
 * We use a tight protocol now, completely doing each outport one
 * at a time.
 *
 * We cannot, however, listen only to the connection that we
 * create for this migration, since an outport migration request may
 * come in in the meantime. We also require that the inport migration ack
 * follow any messages from the other process.
 *
 * Algorithm:
 *   For each outport
 *      Construct and send migration request message. Contains
 *	    current conn/id
 *	    new conn/id
 *      Wait for response. We enqueue everything but outport migration
 *	permreqs and inport migration acks.
 */
 
void _fm_migrate_inport(fm_inport_t inport, conn_t new_conn, int new_id)
{
    fm_outport_t outport;
    fm_outport_reference_t *outport_ref;
    int n_outports;
    bool_t *closed_outports;
    int i;


#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("fm_migrate_inport(): inport=%s new_conn=%s new_id=%d\n",
		    _fm_inport_as_string(inport),
		    yoke_conn_as_string(new_conn),
		    new_id);
#endif
    
    _fm_process_get_closed_info(inport->channel_id, &n_outports, &closed_outports);
    Assert(n_outports == inport->n_outports);

    for (i = 0; i < inport->n_outports; i++)
    {
	outport_ref = &(inport->outports[i]);

	if (closed_outports[i])
	    continue;

	if (outport_ref->local_outport != (fm_outport_t) NULL)
	{
	    notify_local_outport(inport, new_conn, new_id,
				 outport_ref->local_outport);
	}
	else
	{
	    notify_remote_outport(inport, i,
				  new_conn, new_id,
				  outport_ref->remote_outport);
	}
    }
}

static void notify_local_outport(fm_inport_t inport,
				 conn_t new_conn, int new_id,
				 fm_outport_t outport)
{
#ifdef DEBUG
    if (MigrateDebug(1))
	yoke_printf("notify_local_outport(): inport=%s new_conn=%s new_id=%d outport=%s\n",
		    _fm_inport_as_string(inport),
		    yoke_conn_as_string(new_conn),
		    new_id,
		    _fm_outport_as_string(outport));
#endif
    _fm_pindex_remove_outport(outport);
    outport->channel_conn = new_conn;
    outport->channel_id = new_id;
    _fm_pindex_add_outport(outport);
}
				 
static void notify_remote_outport(fm_inport_t inport, int inport_index,
				 conn_t new_conn, int new_id,
				 conn_t outport_conn)
{
    yoke_buffer_t buf;
    bool_t was_established;
    conn_t sender;
    int type;
    bool_t done;
    
#ifdef DEBUG
    if (MigrateDebug(1))
	yoke_printf("notify_remote_outport(): inport=%s new_conn=%s new_id=%d outport_conn=%s\n",
		    _fm_inport_as_string(inport),
		    yoke_conn_as_string(new_conn),
		    new_id,
		    yoke_conn_as_string(outport_conn));
#endif

    /*
     * Construct the migration request
     */

    buf = yoke_allocate_buffer(outport_conn,
			       yoke_sizeof_conn(outport_conn, inport->channel_conn) +
			       yoke_sizeof_conn(outport_conn, new_conn) +
			       yoke_sizeof_int(outport_conn, 3),
			       5);
    yoke_put_conn(buf, inport->channel_conn);
    yoke_put_int(buf, &inport->channel_id, 1);
    yoke_put_conn(buf, new_conn);
    yoke_put_int(buf, &new_id, 1);
    yoke_put_int(buf, &inport_index, 1);

    if (!(was_established = yoke_is_established(outport_conn)))
	yoke_establish(outport_conn);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("notify_remote_outport: send INPORT_MIGRATE to %s\n",
		    yoke_conn_as_string(outport_conn));
#endif
    yoke_send(outport_conn, buf, FM_MTYPE_INPORT_MIGRATE);

    /*
     * We need only signal if we weren't already connected.
     */
    if (!was_established)
    {
	conn_t mgr;
	
	yoke_release(outport_conn);

	_fm_signal_conn("notify_remote_outport()", outport_conn);
    }

    /*
     * Wait for the ack to come back. If we get a permission request,
     * just ignore it. Any other message (other than the ack) is
     * enqueued for later processing.
     *
     */

    done = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: notify_remote_outport() waiting for ack\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_INPORT_MIGRATE_ACK)
	{
	    /*
	     * We're done. (This is now an empty message).
	     */
	    yoke_free_buffer(buf);
	    done = TRUE;
	    yoke_release_transient(sender);
	}
	else if (type == FM_MTYPE_OUTPORT_MIGRATE)
	{
	    conn_t reply_conn, channel_conn;
	    int channel_id;
	    
	    /*
	     * Ignore the message if it is for this inport.
	     */
	    yoke_get_conn(buf, &reply_conn);
	    yoke_get_conn(buf, &channel_conn);
	    yoke_get_int(buf, &channel_id, 1);

	    if (channel_id == inport->channel_id)
	    {
		yoke_free_buffer(buf);
	    }
	    else
	    {
		yoke_reset_buffer(buf);
		_fm_do_outport_migration(sender, buf, TRUE);
	    }
	    yoke_release_transient(sender);
	}
#ifdef OLD_CODE
	else if (type == FM_MTYPE_OUTPORT_PERM_REQ)
	{
	    /*
	     * Ignore it if it is for this inport.
	     */
	    int channel_id;

	    yoke_get_int(buf, &channel_id, 1);
	    if (channel_id == inport->channel_id)
	    {
		yoke_free_buffer(buf);
	    }
	    else
	    {
		/*
		 * Otherwise go ahead and process it.
		 */
/**/
		yoke_reset_buffer(buf);
#if 0
		_fm_process_permission_req(buf, FALSE);
#endif
		done = process_permission_req_checking_for_inport_migrate_ack(buf, FALSE);
/**/
/*
		yoke_reset_buffer(buf);
		_fm_enqueue_migration_request(buf, type);
*/
	    }
	    yoke_release_transient(sender);
	}
#endif /* OLD_CODE */
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
	}
    }

    /*
     * Handle anything that we deferred.
     */

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("notify_remote_outport: handling queued requests\n");
#endif
     _fm_handle_queued_migration_requests();
}


void _fm_do_inport_migration(conn_t sender, yoke_buffer_t buf, bool_t send_ack)
{
    conn_t old_conn, new_conn;
    int old_id, new_id;
    fm_inport_t inport;
    fm_outport_t outport, *outports;
    int n_outports, i, j, inport_index;

    yoke_get_conn(buf, &old_conn);
    yoke_get_int(buf, &old_id, 1);
    yoke_get_conn(buf, &new_conn);
    yoke_get_int(buf, &new_id, 1);
    yoke_get_int(buf, &inport_index, 1);

    do_inport_migration_guts(old_conn, old_id, new_conn, new_id, inport_index, send_ack);
}

static void do_inport_migration_guts(conn_t old_conn, int old_id,
				     conn_t new_conn, int new_id,
				     int inport_index, bool_t send_ack)
{
    yoke_buffer_t buf;
    fm_inport_t inport;
    fm_outport_t outport, *outports;
    int n_outports, i, j;

#ifdef DEBUG
    if (MigrateDebug(2))
    {
	yoke_printf("%s: _fm_do_inport_migration(): Got old_conn=%s old_id=%d\n",
		    _fm_process_name,
		    yoke_conn_as_string(old_conn), old_id);
	yoke_printf("%s: _fm_do_inport_migration(): Got new_conn=%s new_id=%d\n inport_index=%d\n",
		    _fm_process_name,
		    yoke_conn_as_string(new_conn), new_id,
		    inport_index);
    }
#endif

    _fm_pindex_lookup_ports(old_conn, old_id, &inport, &outports, &n_outports);

#ifdef DEBUG
    if (inport != (fm_inport_t) NULL)
    {
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Got inport %s\n",
			_fm_process_name,
			_fm_inport_as_string(inport));
    }
#endif

    /*
     * Update the outport data structures.
     */
    
    if (outports != (fm_outport_t *) NULL)
    {
	bool_t gotit;
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Got %d outports\n",
			_fm_process_name,
			n_outports);
#endif
	gotit = FALSE;
	for (i = 0; i < n_outports; i++)
	{
	    outport = outports[i];

	    if (outport->inport_index == inport_index)
	    {
		
		/*
		 * Got the outport in question. Drop out of the loop.
		 */
		gotit = TRUE;
		break;
	    }
	}

	if (!gotit)
	{
	    yoke_fatal("%s: _fm_do_inport_migration(): Didn't find outport for conn=%s id=%d inport_index=%d (new_conn=%s new_id=%d)\n",
		       _fm_process_name,
		       yoke_conn_as_string(old_conn), old_id,
		       inport_index,
		       yoke_conn_as_string(new_conn), new_id);
	}

	Assert(yoke_conn_equal(outport->channel_conn, old_conn));
	Assert(outport->channel_id == old_id);

	_fm_pindex_remove_outport(outport);
	outport->channel_conn = new_conn;
	outport->channel_id = new_id;
	_fm_pindex_add_outport(outport);
    }
    else
    {
	yoke_fatal("%s: _fm_do_inport_migration(): Didn't find outport for conn=%s id=%d inport_index=%d (new_conn=%s new_id=%d)\n",
		   _fm_process_name,
		   yoke_conn_as_string(old_conn), old_id,
		   inport_index,
		   yoke_conn_as_string(new_conn), new_id);
    }


    if (send_ack)
    {
	buf = yoke_allocate_buffer(old_conn, 0, 0);
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_do_inport_migration(): Sending ack to old_conn=%s\n",
			_fm_process_name,
			yoke_conn_as_string(old_conn));
#endif
	yoke_establish(old_conn);
	yoke_send(old_conn, buf, FM_MTYPE_INPORT_MIGRATE_ACK);
	yoke_release(old_conn);
    }
}


/*
 * _fm_do_outport_migration()
 *
 * Process an outport migration request. sender and buf are as
 * returned from yoke_recv_transient(). Do not invoke
 * yoke_release_transient() on the conn_t as the sender is
 * responsible for doing that.
 */
void _fm_do_outport_migration(conn_t sender, yoke_buffer_t buf, bool_t send_ack)
{
    conn_t channel_conn, new_conn, reply_conn;
    int inport_index;
    int channel_id;
    
    yoke_get_conn(buf, &reply_conn);
    yoke_get_conn(buf, &channel_conn);
    yoke_get_int(buf, &channel_id, 1);
    yoke_get_int(buf, &inport_index, 1);
    yoke_get_conn(buf, &new_conn);

    _fm_do_outport_migration_guts(reply_conn,
				  channel_conn,
				  channel_id,
				  inport_index,
				  new_conn, send_ack);
}

static void _fm_do_outport_migration_guts(conn_t reply_conn,
				   conn_t channel_conn,
				   int channel_id,
				   int inport_index,
				   conn_t new_conn, bool_t send_ack)
{
    yoke_buffer_t buf;
    fm_inport_t inport;
    fm_outport_t outport, *outports;
    fm_outport_reference_t *ref;
    int n_outports, i;
 
#ifdef DEBUG
    if (MigrateDebug(2))
    {
	yoke_printf("%s: _fm_do_outport_migration(): Got channel_conn=%s channel_id=%d inport_index=%d new_conn=%s\n",
		    _fm_process_name,
		    yoke_conn_as_string(channel_conn),
		    channel_id,
		    inport_index,
		    yoke_conn_as_string(new_conn));
    }
#endif

    _fm_pindex_lookup_ports(channel_conn, channel_id,
			    &inport, &outports, &n_outports);

    if (inport != (fm_inport_t) NULL)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("_fm_do_outport_migration(): Got inport %s\n",
			_fm_inport_as_string(inport));
#endif
	Assert(inport_index >= 0 && inport_index < inport->n_outports);
	ref = &(inport->outports[inport_index]);
	Assert(ref->local_outport == (fm_outport_t) NULL);
	Assert(yoke_conn_equal(ref->remote_outport, reply_conn));

	ref->remote_outport = new_conn;
    }
    else
    {
	yoke_fatal("_fm_do_outport_migration(): Didn't find the inport for channel%s id=%d\n",
		   yoke_conn_as_string(channel_conn), channel_id);
    }
    
    if (send_ack)
    {
	buf = yoke_allocate_buffer(reply_conn, 0, 0);

#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("_fm_do_outport_migration(): Sending ack to reply=%s\n",
			yoke_conn_as_string(reply_conn));
#endif
	yoke_establish(reply_conn);
	yoke_send(reply_conn, buf, FM_MTYPE_OUTPORT_MIGRATE_ACK);
	yoke_release(reply_conn);
    }
}


/*
 * _fm_migrate_outport()
 *
 * Migrate the outport to its future home (outport->future_home).
 *
 * Algorithm:
 *
 *	Send OUTPORT_MIGRATE to inport
 *      Wait for messages. If we get a
 *	    OUTPORT_MIGRATE_ACK: it worked
 *	    INPORT_MIGRATE: do the inport migration, send an ack, and resend
 *			    the outport migration request
 *	    Anything else, enqueue until done.
 */
void _fm_migrate_outport(fm_outport_t outport)
{
    int i;
    bool_t done;
    conn_t sender;
    int type;
    yoke_buffer_t buf;
    bool_t was_established;
    conn_t reply_conn;

    Assert(outport->future_home != (conn_t) NULL);

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("fm_migrate_outport(): %s\n",
		    _fm_outport_as_string(outport));
#endif

    /*
     * Send the actual migration request.
     */

    send_outport_migration_request(outport);

    /*
     * And wait for the outport migration acknowledgement.
     */
    done = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_migrate_outport() waiting for ack\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_OUTPORT_MIGRATE_ACK)
	{
	    /*
	     * We're done. (This is now an empty message).
	     */
	    yoke_free_buffer(buf);
	    done = TRUE;
	    yoke_release_transient(sender);
	}
	else if (type == FM_MTYPE_INPORT_MIGRATE)
	{
	    /*
	     * If this is a migration that affects this outport,
	     * process it and resend the migration request.
	     */
	    process_inport_migrate_during_outport_migrate(outport,
							  type, buf);
	}
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
/*
	    _fm_enqueue_migration_request(buf, type);
*/
	}
    }

    /*
     * Delete our entry from the index, since we're officially migrated.
     */
    _fm_pindex_remove_outport(outport);

    /*
     * Handle anything that we deferred.
     */

    _fm_handle_queued_migration_requests();
}

static void send_outport_migration_request(fm_outport_t outport)
{
    conn_t reply_conn;
    yoke_buffer_t buf;
    bool_t was_established;
    
    reply_conn = yoke_self();
    buf = yoke_allocate_buffer(outport->channel_conn,
			       yoke_sizeof_int(outport->channel_conn, 2) +
			       yoke_sizeof_conn(outport->channel_conn,
						outport->future_home) +
			       yoke_sizeof_conn(reply_conn,
						outport->future_home) +
			       yoke_sizeof_conn(outport->channel_conn,
						outport->channel_conn),
			       5);
    yoke_put_conn(buf, reply_conn);
    yoke_put_conn(buf, outport->channel_conn);
    yoke_put_int(buf, &outport->channel_id, 1);
    yoke_put_int(buf, &outport->inport_index, 1);
    yoke_put_conn(buf, outport->future_home);
    if (!(was_established = yoke_is_established(outport->channel_conn)))
	yoke_establish(outport->channel_conn);
    
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("_fm_migrate_outport(): send OUTPORT_MIGRATE to %s\n",
		    yoke_conn_as_string(outport->channel_conn));
#endif
    yoke_send(outport->channel_conn, buf, FM_MTYPE_OUTPORT_MIGRATE);

    /*
     * We need only signal if we weren't already connected.
     */
    if (!was_established)
    {
	yoke_release(outport->channel_conn);
	_fm_signal_conn("send_outport_migration_request()", outport->channel_conn);
    }
}

#if 0 /* OLD CODE */

/*
 * _fm_migrate_outport()
 *
 * Migrate the outport to its future home (outport->future_home).
 *
 * Algorithm:
 *
 *	Send OUTPORT_PERM_REQ to inport
 *      Wait for messages. If we get a
 *          PERM_GRANTED: send real request and wait for ack
 *	    INPORT_MIGRATE: do the inport migration, send an ack, and resend
 *			    the outport migration request
 *	    Anything else, enqueue until done.
 */
void _fm_migrate_outport(fm_outport_t outport)
{
    int i;
    bool_t done;
    conn_t sender;
    int type;
    yoke_buffer_t buf;
    bool_t was_established;
    conn_t reply_conn;

    Assert(outport->future_home != (conn_t) NULL);

    /*
     * Send a permission request.
     */

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("fm_migrate_outport(): %s\n",
		    _fm_outport_as_string(outport));
#endif

    reply_conn = yoke_self();
    buf = yoke_allocate_buffer(outport->channel_conn,
			       yoke_sizeof_int(outport->channel_conn, 1) +
			       yoke_sizeof_conn(outport->channel_conn, reply_conn),
			       2);
    yoke_put_int(buf, &outport->channel_id, 1);
    yoke_put_conn(buf, reply_conn);

    if (!(was_established = yoke_is_established(outport->channel_conn)))
	yoke_establish(outport->channel_conn);
    
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("%s: _fm_migrate_outport(): sending first OUTPORT_PERM_REQ to %s\n",
		    _fm_process_name,
		    yoke_conn_as_string(outport->channel_conn));
#endif
    yoke_send(outport->channel_conn, buf, FM_MTYPE_OUTPORT_PERM_REQ);

    /*
     * We need only signal if we weren't already connected.
     */
    if (!was_established)
    {
	yoke_release(outport->channel_conn);
	_fm_signal_conn("_fm_migrate_outport()", outport->channel_conn);
    }

    /*
     * Wait for a reply to the permission request.
     */

    done = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_migrate_outport() waiting for permission reply\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_INPORT_MIGRATE)
	{
	    /*
	     * We need to look at the buffer to decide if this is
	     * an inport migration for the outport we are currently
	     * trying to migrate. If it is, go ahead and perform the
	     * migration and resend the permission request.
	     */

	    process_inport_migrate_during_outport_migrate(outport, type, buf);
	}
	else if (type == FM_MTYPE_OUTPORT_PERM_GRANTED)
	{
	    /*
	     * We're on. Drop out of the loop and go ahead
	     * with the migration.
	     */
	    yoke_free_buffer(buf);
	    done = TRUE;
	}
	else if (type == FM_MTYPE_OUTPORT_PERM_REQ)
	{
	    /*
	     * Queue this so we don't get confused.
	     */
	    _fm_enqueue_migration_request(buf, type);
	}
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
/*
	    _fm_enqueue_migration_request(buf, type);
*/
	}
	yoke_release_transient(sender);
    }

    /*
     * We can go ahead with the standard migration method.
     * Send the actual migration request.
     */
    reply_conn = yoke_self();
    buf = yoke_allocate_buffer(outport->channel_conn,
			       yoke_sizeof_int(outport->channel_conn, 2) +
			       yoke_sizeof_conn(outport->channel_conn,
						outport->future_home) +
			       yoke_sizeof_conn(reply_conn,
						outport->future_home) +
			       yoke_sizeof_conn(outport->channel_conn,
						outport->channel_conn),
			       5);
    yoke_put_conn(buf, reply_conn);
    yoke_put_conn(buf, outport->channel_conn);
    yoke_put_int(buf, &outport->channel_id, 1);
    yoke_put_int(buf, &outport->inport_index, 1);
    yoke_put_conn(buf, outport->future_home);
    if (!(was_established = yoke_is_established(outport->channel_conn)))
	yoke_establish(outport->channel_conn);
    
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("_fm_migrate_outport(): send OUTPORT_MIGRATE to %s\n",
		    yoke_conn_as_string(outport->channel_conn));
#endif
    yoke_send(outport->channel_conn, buf, FM_MTYPE_OUTPORT_MIGRATE);

    /*
     * We need only signal if we weren't already connected.
     */
    if (!was_established)
    {
	yoke_release(outport->channel_conn);
	_fm_signal_conn("_fm_migrate_outport()", outport->channel_conn);
    }

    /*
     * And wait for the outport migration acknowledgement.
     */
    done = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_migrate_outport() waiting for ack\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_OUTPORT_MIGRATE_ACK)
	{
	    /*
	     * We're done. (This is now an empty message).
	     */
	    yoke_free_buffer(buf);
	    done = TRUE;
	    yoke_release_transient(sender);
	}
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
/*
	    _fm_enqueue_migration_request(buf, type);
*/
	}
    }

    /*
     * Delete our entry from the index, since we're officially migrated.
     */
    _fm_pindex_remove_outport(outport);

    /*
     * Handle anything that we deferred.
     */

    _fm_handle_queued_migration_requests();
}

#endif /* OLD CODE */

static void process_inport_migrate_during_outport_migrate(fm_outport_t outport,
							  int type,
							  yoke_buffer_t buf)
{
    conn_t old_conn, new_conn, reply_conn;
    int old_id, new_id;
    int inport_index;
    bool_t was_established;

    yoke_get_conn(buf, &old_conn);
    yoke_get_int(buf, &old_id, 1);
    yoke_get_conn(buf, &new_conn);
    yoke_get_int(buf, &new_id, 1);
    yoke_get_int(buf, &inport_index, 1);


    if (yoke_conn_equal(outport->channel_conn, old_conn) &&
	outport->channel_id == old_id &&
	outport->inport_index == inport_index)
    {
	/*
	 * This is a migration for the outport we are currently migrating.
	 * Go ahead and do the migration, and resend the request.
	 */
	yoke_free_buffer(buf);
	do_inport_migration_guts(old_conn, old_id, new_conn, new_id,
				 inport_index, TRUE);

	/*
	 * ... resend the request ...
	 */

	send_outport_migration_request(outport);
    }
    else
    {
	/*
	 * This is some other migration. Enqueue it.
	 */

	yoke_reset_buffer(buf);
	_fm_enqueue_migration_request(buf, type);
    }
}


#if 0 /* OLD CODE */

static void process_inport_migrate_during_outport_migrate(fm_outport_t outport,
							  int type,
							  yoke_buffer_t buf)
{
    conn_t old_conn, new_conn, reply_conn;
    int old_id, new_id;
    int inport_index;
    bool_t was_established;

    yoke_get_conn(buf, &old_conn);
    yoke_get_int(buf, &old_id, 1);
    yoke_get_conn(buf, &new_conn);
    yoke_get_int(buf, &new_id, 1);
    yoke_get_int(buf, &inport_index, 1);


    if (yoke_conn_equal(outport->channel_conn, old_conn) &&
	outport->channel_id == old_id &&
	outport->inport_index == inport_index)
    {
	/*
	 * This is a migration for the outport we are currently migrating.
	 * Go ahead and do the migration, and resend the request.
	 */
	yoke_free_buffer(buf);
	do_inport_migration_guts(old_conn, old_id, new_conn, new_id,
				 inport_index, TRUE);

	/*
	 * ... resend the request ...
	 */
	
	reply_conn = yoke_self();
	buf = yoke_allocate_buffer(outport->channel_conn,
				   yoke_sizeof_int(outport->channel_conn, 1) +
				   yoke_sizeof_conn(outport->channel_conn, reply_conn),
				   2);
	yoke_put_int(buf, &outport->channel_id, 1);
	yoke_put_conn(buf, reply_conn);

	if (!(was_established = yoke_is_established(outport->channel_conn)))
	    yoke_establish(outport->channel_conn);
    
#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("%s: process_inport_migrate_during_outport_migrate(): sending OUTPORT_PERM_REQ to %s\n",
		    _fm_process_name,
		    yoke_conn_as_string(outport->channel_conn));
#endif
	yoke_send(outport->channel_conn, buf, FM_MTYPE_OUTPORT_PERM_REQ);
	if (!was_established)
	{
	    yoke_release(outport->channel_conn);
	    _fm_signal_conn("process_inport_migrate_during_outport_migrate()", outport->channel_conn);
	}
    }
    else
    {
	/*
	 * This is some other migration. Enqueue it.
	 */

	yoke_reset_buffer(buf);
	_fm_enqueue_migration_request(buf, type);
    }
}

#endif /* OLD CODE */


/*
 * _fm_process_permission_req()
 *
 * Process a FM_MTYPE_OUTPORT_PERM_REQ that is received
 * during normal operation (i.e. not during an inport
 * migration). Return a FM_MTYPE_OUTPORT_PERM_GRANTED
 * message and loop waiting for the actual request,
 * enqueuing messages until it is received.
 */
void _fm_process_permission_req(yoke_buffer_t in_buf,
				bool_t process_queue)
{									
    conn_t reply_conn;							
    int channel_id;							
    yoke_buffer_t buf;
    int type;
    conn_t sender;
    bool_t done;
    
    conn_t r_channel_conn, r_new_conn, r_reply_conn;
    int r_inport_index;
    int r_channel_id;
    
    yoke_get_int(in_buf, &channel_id, 1);					
    yoke_get_conn(in_buf, &reply_conn);					

    buf = yoke_allocate_buffer(reply_conn, 0, 0);			
    yoke_establish(reply_conn);					
    yoke_send(reply_conn, buf, FM_MTYPE_OUTPORT_PERM_GRANTED);	
    yoke_release(reply_conn);

    done = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_process_permission_req() waiting for actual migration\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_OUTPORT_MIGRATE)
	{
	    yoke_get_conn(buf, &r_reply_conn);
	    yoke_get_conn(buf, &r_channel_conn);
	    yoke_get_int(buf, &r_channel_id, 1);
	    yoke_get_int(buf, &r_inport_index, 1);
	    yoke_get_conn(buf, &r_new_conn);

	    if (r_channel_id == channel_id)
	    {
		/*
		 * OK, this is it.
		 * Do the migration and drop out of the loop
		 */
		_fm_do_outport_migration_guts(r_reply_conn,
					      r_channel_conn,
					      r_channel_id,
					      r_inport_index,
					      r_new_conn, TRUE);
		yoke_free_buffer(buf);
		done = TRUE;
	    }
	    else
	    {
		/*
		 * Nope, we're not interested.
		 */
		yoke_reset_buffer(buf);
		_fm_handle_message(sender, type, buf, FALSE);
	    }
	}
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
	}
	yoke_release_transient(sender);
    }

    /*
     * Handle anything that we deferred.
     */

    if (process_queue)
	_fm_handle_queued_migration_requests();
}

/*
 * Yuk what a hack. The same as above, but returns true if
 * it received an inport migration ack.
 */
static bool_t process_permission_req_checking_for_inport_migrate_ack
(yoke_buffer_t in_buf, bool_t process_queue)
{									
    conn_t reply_conn;							
    int channel_id;							
    yoke_buffer_t buf;
    int type;
    conn_t sender;
    bool_t done;
    bool_t got_ack;
    
    conn_t r_channel_conn, r_new_conn, r_reply_conn;
    int r_inport_index;
    int r_channel_id;
    
    yoke_get_int(in_buf, &channel_id, 1);					
    yoke_get_conn(in_buf, &reply_conn);					

    buf = yoke_allocate_buffer(reply_conn, 0, 0);			
    yoke_establish(reply_conn);					
    yoke_send(reply_conn, buf, FM_MTYPE_OUTPORT_PERM_GRANTED);	
    yoke_release(reply_conn);

    done = FALSE;
    got_ack = FALSE;
    while (!done)
    {
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("%s: _fm_process_permission_req() waiting for actual migration\n",
			_fm_process_name);
#endif

	buf = yoke_recv_transient(&sender, &type, FALSE);

#ifdef DEBUG
	if (MigrateDebug(3))
	    yoke_printf("Got message type=%d %s sender=%s\n",
			type, _fm_mtype_name(type),
			yoke_conn_as_string(sender));
#endif

	if (type == FM_MTYPE_OUTPORT_MIGRATE)
	{
	    yoke_get_conn(buf, &r_reply_conn);
	    yoke_get_conn(buf, &r_channel_conn);
	    yoke_get_int(buf, &r_channel_id, 1);
	    yoke_get_int(buf, &r_inport_index, 1);
	    yoke_get_conn(buf, &r_new_conn);

	    if (r_channel_id == channel_id)
	    {
		/*
		 * OK, this is it.
		 * Do the migration and drop out of the loop
		 */
		_fm_do_outport_migration_guts(r_reply_conn,
					      r_channel_conn,
					      r_channel_id,
					      r_inport_index,
					      r_new_conn, TRUE);
		yoke_free_buffer(buf);
		done = TRUE;
	    }
	    else
	    {
		/*
		 * Nope, we're not interested.
		 */
		yoke_reset_buffer(buf);
		_fm_handle_message(sender, type, buf, FALSE);
	    }
	}
	else if (type == FM_MTYPE_INPORT_MIGRATE_ACK)
	{
	    got_ack = TRUE;
	    yoke_free_buffer(buf);
	}
	else
	{
	    _fm_handle_message(sender, type, buf, FALSE);
	}
	yoke_release_transient(sender);
    }

    /*
     * Handle anything that we deferred.
     */

    if (process_queue)
	_fm_handle_queued_migration_requests();

    return got_ack;
}

void _fm_signal_conn(char *func, conn_t conn)
{
    conn_t mgr;
    yoke_buffer_t buf;

    mgr = _fm_get_manager_for_conn(conn);

    buf = yoke_allocate_buffer(mgr,
			       yoke_sizeof_conn(mgr, conn),
			       1);
    yoke_put_conn(buf, conn);
    yoke_establish(mgr);
    yoke_send(mgr, buf, FM_SIGNAL_CHILD);

    yoke_release(mgr);
}
