/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"


static char msg_buf[MAX_ABEND_MSG_SIZE+1];

/*
 * FM_SHUTDOWN()
 *
 * Shut down the FM computation. Any node process can invoke this
 * routine. The protocol is
 *
 *    Send SHUTDOWN_REQUEST to manager and exit.
 *    If manager is not manager 0, forwards request to manager 0
 *    Manager 0 sends MGR_SHUTDOWN_1 to all managers, leaving
 *      connections established.
 *    Managers respond with MGR_SHUTDOWN_1_ACK
 *    Managers kill off any remaining children
 *    Manager 0 sends MGR_SHUTDOWN_2 to all managers and releases connection.
 *    Managers respond to MGR_SHUTDOWN_2 by exiting.
 *    
 */
void FM_SHUTDOWN()
{
    yoke_buffer_t buf;
    conn_t conn;
    int type, pid;

#ifdef DEBUG
    if (FMDebug(3))
	yoke_printf("FM shutdown\n");
#endif

    StartSafeRegion;

    buf = yoke_allocate_buffer(_fm_my_manager, yoke_sizeof_int(_fm_my_manager, 1), 1);
    pid = getpid();
    yoke_put_int(buf, &pid, 1);
    yoke_establish(_fm_my_manager);
    yoke_send(_fm_my_manager, buf, FM_MTYPE_SHUTDOWN_REQUEST);
    buf = yoke_recv(_fm_my_manager, &type);
    if (type != FM_MTYPE_SHUTDOWN_REQUEST_ACK)
	yoke_fatal("FM_SHUTDOWN(): expected fm_MTYPE_SHUTDOWN_REQUEST_ACK, got %s %d\n",
		   _fm_mtype_name(type),type);
    yoke_free_buffer(buf);
    yoke_release(_fm_my_manager);

    yoke_shutdown();
    exit(0);
}


/*
 * _fm_handle_shutdown_request()
 *
 * Manager 0 executes this. Broadcast a MGR_SHUTDOWN_1 to each manager, 
 * wait for an ack, and then broadcast a MGR_SHUTDOWN_2 and exit.
 */
void _fm_handle_shutdown_request()
{
    int i, type;
    yoke_buffer_t buf;
    conn_t conn;
    int n_replies, who;

    for (i = 1; i < _fm_n_managers(); i++)
    {
	conn = _fm_get_manager(i);

	buf = yoke_allocate_buffer(conn, 0, 0);
	yoke_establish(conn);
#ifdef DEBUG
	    if (FMDebug(3))
		yoke_printf("_fm_handle_shutdown_request(): sending SHUTDOWN_1 to %d\n", i);
#endif
	yoke_send(conn, buf, FM_MTYPE_MGR_SHUTDOWN_1);
    }

    _fm_manager_clean_up_children();
    
    for (i = 1; i < _fm_n_managers(); i++)
    {
	conn = _fm_get_manager(i);
	buf = yoke_recv(conn, &type);
	if (type != FM_MTYPE_MGR_SHUTDOWN_1_ACK)
	{
	    yoke_fatal("_fm_handle_shutdown_request(): got type %s %d instead of QUIT_ACK\n",
		       _fm_mtype_name(type), type);
	}
    }

    for (i = 1; i < _fm_n_managers(); i++)
    {
	conn = _fm_get_manager(i);

	buf = yoke_allocate_buffer(conn, 0, 0);
#ifdef DEBUG
	    if (FMDebug(3))
		yoke_printf("_fm_handle_shutdown_request(): sending SHUTDOWN_2 to %d\n", i);
#endif
	yoke_send(conn, buf, FM_MTYPE_MGR_SHUTDOWN_2);
    }

    yoke_shutdown();
#ifdef DEBUG
    if (FMDebug(3))
	yoke_printf("_fm_handle_shutdown_request(): exiting\n");
#endif
    exit(0);
}


/*
 * _fm_handle_abend()
 *
 * Handle an abnormal termination. Manager 0 invokes
 * this upon receipt of a FM_MTYPE_ABEND from
 * another manager who has noted the abnormal termination
 * of a child process.
 */
void _fm_handle_abend(yoke_buffer_t buf)
{
    int msgsize;

    yoke_get_int(buf, &msgsize, 1);
    yoke_get_char(buf, msg_buf, msgsize);
    yoke_printf("Abnormal termination: %s\n", msg_buf);
    _fm_handle_shutdown_request();
}

/*
 * _fm_shutdown_one_manager()
 *
 * Code executed by managers 1..n  upon receipt of
 * MGR_SHUTDOWN_1.
 */
void _fm_shutdown_one_manager(conn_t sender)
{
    yoke_buffer_t buf;
    int type;
    bool_t got_shutdown_2;

    buf = yoke_allocate_buffer(sender, 0, 0);
    yoke_send(sender, buf, FM_MTYPE_MGR_SHUTDOWN_1_ACK);

    got_shutdown_2 = _fm_manager_clean_up_children();

    if (!got_shutdown_2)
    {
	buf = yoke_recv(sender, &type);

	if (type != FM_MTYPE_MGR_SHUTDOWN_2)
	    yoke_fatal("_fm_shutdown_one_manager(): Got type %s %d instead of FM_MTYPE_MGR_SHUTDOWN_2\n",
		       _fm_mtype_name(type), type);
    }
    yoke_shutdown();
    exit(0);
}
