/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "sys_incs.h"

#include <pwd.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/time.h>

#ifdef SIGUSR1
#define NOTIFICATION_SIGNAL SIGUSR1
#endif

#ifndef YOKE_SIGNAL_RETURN_T
#define YOKE_SIGNAL_RETURN_T void
#endif

#include "yoke_internal.h"

/*
 * Command line arguments...
 */
static bool_t   trapping_sigtrap;
static bool_t   trapping_sigfpe;
static bool_t	no_trapping;

static void child_death(int s);
static void abnormal_death(int s);

static char short_hostname_id[MAXHOSTNAMELEN];
static char hostname_id[MAXHOSTNAMELEN+10];
static int  process_id;

static void child_death(int s);
static void abnormal_death(int s);
static void interrupt(int s);
static void terminate(int s);

static char short_hostname_id[MAXHOSTNAMELEN];
static char hostname_id[MAXHOSTNAMELEN+10];
static int  process_id;

static void (*child_death_func)(int, int, int);
static void (*notification_func)();
static YOKE_SIGNAL_RETURN_T notify_handler(int);
static YOKE_SIGNAL_RETURN_T fatal_notify_handler(int s);

static int saved_sigmask;
static bool_t notification_blocked;


#ifdef MALLOC_DEBUG

#ifdef NeXT

#define SetupMallocConsistencyCheckGuts \
	malloc_debug(31); \
	malloc_error(malloc_error_func);
static void malloc_error_func(int);

#endif /* NeXT */

#ifdef sun
#define SetupMallocConsistencyCheckGuts \
	malloc_debug(2);
#endif /* sun */

#ifndef SetupMallocConsistencyCheckGuts
#define SetupMallocConsistencyCheckGuts
#endif /* !SetupMallocConsistencyCheckGuts */

#define SetupMallocConsistencyCheck \
	if (YokeMallocDebug(1)) \
	{ \
	    SetupMallocConsistencyCheckGuts;\
        }

#else /* MALLOC_DEBUG */

#define SetupMallocConsistencyCheck

#endif /* MALLOC_DEBUG */


/*
 * _y_md_init()
 *
 * Initialize any machine dependent stuff
 */
void _y_md_init()
{
    char *s;
    char tmp[MAXHOSTNAMELEN];

    _y_dont_invoke_child_death_notification = FALSE;
    _y_received_child_death_signal = FALSE;
    
    signal(SIGCHLD, child_death);
    signal(SIGINT, interrupt);
    signal(SIGTERM, terminate);
    signal(SIGPIPE, SIG_IGN);
    if (!no_trapping)
    {
	signal(SIGBUS, abnormal_death);
	signal(SIGSEGV, abnormal_death);
	signal(SIGABRT, abnormal_death);
	signal(SIGILL, abnormal_death);
    }
    signal(NOTIFICATION_SIGNAL, notify_handler);

    SetupMallocConsistencyCheck;

    notification_blocked = FALSE;

    process_id = getpid();

    _y_md_gethostname(hostname_id, sizeof(hostname_id));
    
    strcpy(tmp, hostname_id);
    if ((s = index(tmp, '.')) != (char *) NULL)
	*s = '\0';
    sprintf(short_hostname_id, "%s:%d", tmp, process_id);

    notification_func = (void (*)()) NULL;
    child_death_func = (void (*)(int, int, int)) NULL;
} /* _y_md_init() */


/*
 * _y_md_reinit()
 *
 * Reinitialize any machine dependent stuff.  This is called by yoke_fork().
 */
void _y_md_reinit()
{
    char *s;
    char tmp[MAXHOSTNAMELEN];
    
    _y_dont_invoke_child_death_notification = FALSE;
    _y_received_child_death_signal = FALSE;
    
    SetupMallocConsistencyCheck;

    notification_blocked = FALSE;

    notification_func = (void (*)()) NULL;
    child_death_func = (void (*)(int, int, int)) NULL;

    _y_md_gethostname(hostname_id, sizeof(hostname_id));

    process_id = getpid();
    strcpy(tmp, hostname_id);
    if ((s = index(tmp, '.')) != (char *) NULL)
	*s = '\0';
    sprintf(short_hostname_id, "%s:%d", tmp, process_id);
} /* _y_md_reinit() */


/*
 * _y_md_shutdown()
 *
 * Shutdown any machine dependent stuff.  This is called by yoke_shutdown().
 */
void _y_md_shutdown()
{
    exit(0);
} /* _y_md_shutdown() */


/*
 * _y_md_exit()
 */
void _y_md_exit(int return_code)
{
    exit(return_code);
} /* _y_md_exit() */


/*
 * _y_md_process_arguments_init()
 *
 * Initialize any variables that will be used for processing arguments.
 */
void _y_md_process_arguments_init()
{
    no_trapping = FALSE;
    trapping_sigtrap = FALSE;
    trapping_sigfpe = FALSE;
} /* _y_md_process_arguments_init() */


/*
 * _y_md_process_arguments()
 *
 * Process command line arguments for this module.
 */
int _y_md_process_arguments(int current_arg, int arg_count)
{
    char *arg;

    arg = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg);

    if (strcmp(arg, "-catch_sigtrap") == 0)
    {
	signal(SIGTRAP, abnormal_death);
	trapping_sigtrap = TRUE;
	current_arg++;
    }
    else if (strcmp(arg, "-catch_fpe") == 0)
    {
	signal(SIGFPE, abnormal_death);
	trapping_sigfpe = TRUE;
	current_arg++;
    }
    else if (strcmp(arg, "-no_catching") == 0)
    {
	no_trapping = TRUE;
	current_arg++;
    }
    
    return (current_arg);
} /* _y_md_process_arguments() */


/*
 * _y_md_usage_message()
 */
void _y_md_usage_message()
{
    printf("    -no_catching              : Do not catch any signals\n");
    printf("    -catch_sigtrap            : Catch (and terminate on) the SIGTRAP signal\n");
    printf("    -catch_fpe                : Catch (and terminate on) floating point\n");
    printf("                                exceptions\n");
    
} /* _y_md_usage_message() */


/*
 * _y_md_remote_start_args()
 */
int _y_md_remote_start_args(char *buf, int size)
{
    char tmp_buf1[1024];
    int n_added;
    
    tmp_buf1[0] = '\0';

    if (trapping_sigtrap)
	strcat(tmp_buf1, "-catch_sigtrap ");
    if (trapping_sigfpe)
	strcat(tmp_buf1, "-catch_fpe ");
    if (no_trapping)
	strcat(tmp_buf1, "-no_catching ");

    n_added = strlen(tmp_buf1);
    if (n_added > size)
	yoke_fatal("_y_md_remote_start_args(): Internal error: Not enough room in buffer for arguments\n");

    strcpy(buf, tmp_buf1);
    
    return (n_added);
    
} /* _y_md_remote_start_args() */


/*
 * _y_md_system_error_string()
 *
 * Return the string for the current errno.
 */
char *_y_md_system_error_string()
{
    extern int errno;
    extern char *sys_errlist[];
    
    return sys_errlist[errno];
} /* _y_md_system_error_string() */


/*
 * _y_md_gethostname()
 *
 * Return the hostname that this process is running on.
 * Do any error checking.
 * Map spnode -> swnode on the sp1.
 */
void _y_md_gethostname(char *name, int len)
{
    if (gethostname(name, len) < 0)
	yoke_fatal("_y_md_gethostname(): gethostname() failed\n");

#ifdef rs6000    
    if (_y_anl_sp1_mapping)
    {
	if (strncmp(name, "spnode", 6) == 0)
	    name[1] = 'w';
    }
#endif /* rs6000 */
    
} /* _y_md_gethostname() */


/*
 * _y_md_get_host_id()
 */
void _y_md_get_host_id(char **a_hostname_id,
		       int *a_process_id)
{
    *a_hostname_id = hostname_id;
    *a_process_id = process_id;
} /* _y_md_get_host_id() */


/*
 * _y_md_get_short_host_id()
 */
void _y_md_get_short_host_id(char **a_hostname_id)
{
    *a_hostname_id = short_hostname_id;
} /* _y_md_get_short_host_id() */


/*
 * child_death()
 *
 * Signal handler for SIGCHLD
 */
static void child_death(int s)
{
    if (_y_dont_invoke_child_death_notification)
	_y_received_child_death_signal = TRUE;
    else
	_y_invoke_child_death_notifier();
}


/*
 * _y_invoke_child_death_notifier()
 *
 * Wait for signals and invoke the death notification function.
 * We don't just do this from the signal handler in order to make
 * sure we don't reenter yoke.
 */
void _y_invoke_child_death_notifier()
{
    int pid;
#ifdef YOKE_HAS_UNION_WAIT
    union wait status;
#else
    int status;
#endif
    int signal, retcode;

    _y_received_child_death_signal = FALSE;

#ifdef YOKE_HAS_WAIT3    
    while ((pid = wait3(&status, WNOHANG, NULL)) > 0)
#else /* YOKE_HAS_WAIT3 */

#ifdef YOKE_HAS_WAITPID
    while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
#else /* YOKE_HAS_WAITPID */
#error Need either wait3 or waitpid
#endif /* YOKE_HAS_WAITPID */

#endif /* YOKE_HAS_WAIT3 */
	
    {
#ifdef YOKE_HAS_WAIT_STATUS_MACROS
	signal = WTERMSIG(status);
	retcode = WEXITSTATUS(status);
#else
	signal = status.w_termsig;
	retcode = status.w_retcode;
#endif
	

	if (
#ifdef DEBUG
	    YokeDebug(1) ||
#endif
	    signal > 0 || retcode > 0)
	{
	    yoke_printf("Child %d died with signal=%d retcode=%d\n",
		      pid, signal, retcode);
	}
	if (child_death_func != (void (*)(int, int, int)) NULL)
	    (*child_death_func)(pid, signal, retcode);

    }
} /* child_death() */


/*
 * abnormal_death()
 *
 * Signal handler for various signals that should kill the process
 */
static void abnormal_death(int s)
{
    yoke_fatal("Signal %d\n", s);
} /* abnormal_death() */


/*
 * terminate()
 *
 * Signal handler for SIGTERM, sent by yoke_kill()
 */
static void terminate(int s)
{
    yoke_silent_fatal();
} /* terminate() */


/*
 * interrupt()
 *
 * Signal handler for interrupt.
 */
static void interrupt(int s)
{
    yoke_printf("Exiting due to interrupt\n");
    _y_md_exit(0);
}

/*
 * _y_md_fatal()
 */
void _y_md_fatal(int return_code)
{
    exit(return_code);
} /* _y_md_fatal() */


/*
 * yoke_register_notification_func()
 *
 * Register func as the function to be called when a _y_md_notify()
 * is invoked for that process by another process on the same machine.
 */
void yoke_register_notification_func(void (*func)())
{
    notification_func = func;
} /* yoke_register_notification_func() */


/*
 * yoke_register_child_death_func()
 *
 * Register func as the function to be called when a _y_md_notify()
 * is invoked for that process by another process on the same machine.
 */
void yoke_register_child_death_func(void (*func)(int, int, int))
{
    child_death_func = func;
} /* yoke_register_notification_func() */


/*
 * yoke_notify()
 *
 * Force the notification function on the process referred to by
 * conn to be invoked.
 */
void yoke_notify(conn_t conn)
{
    int pid;
    char *host;

    yoke_get_conn_id(conn, &host, &pid);
    if (strcmp(host, hostname_id) != 0)
	yoke_fatal("yoke_notify(): Cannot notify a process unless it is on the same host.\n");

    /*
     * It is not a fatal error if the process does not exist. This is
     * because the death of that child (if abnormal) will be detected
     * in other ways, and it is possible for a notification to
     * be unnecessary. For instance, if the notification is sent
     * along with an inline message to a process which happens
     * to be waiting for messages, it may process that message
     * and exit normally as a result. Process scheduling may result
     * in the signal being sent after that normal termination.
     */
    if (kill(pid, NOTIFICATION_SIGNAL) != 0)
    {
	if (errno != ESRCH)
	    yoke_fatal_perror("yoke_notify(): Kill failed for %s",
			      yoke_conn_as_string(conn));
    }
    
} /* yoke_notify() */


/*
 * yoke_block_notification()
 *
 * Block invocation of the notification function. If a notify signal comes
 * in, we need to invoke the function after notification has been
 * unblocked.
 */
void yoke_block_notification()
{
    if (notification_blocked)
    {
	yoke_warning("Nested yoke_block_notification()\n");
	return;
    }

#ifdef YOKE_HAS_SIGBLOCK

    saved_sigmask = sigblock(sigmask(NOTIFICATION_SIGNAL));

#else /* YOKE_MAS_SIGBLOCK */

#ifdef YOKE_HAS_SIGHOLD
    
    sighold(NOTIFICATION_SIGNAL);

#else /* YOKE_HAS_SIGHOLD */

#error Must define either YOKE_HAS_SIGBLOCK or YOKE_HAS_SIGHOLD

#endif /* YOKE_HAS_SIGHOLD */
#endif /* YOKE_MAS_SIGBLOCK */

    notification_blocked = TRUE;

} /* yoke_block_notification() */


/*
 * yoke_unblock_notification()
 *
 * Unblock invocation of the notification function.
 */
void yoke_unblock_notification()
{
    if (!notification_blocked)
    {
	yoke_warning("Invoking yoke_unblock_notification() when notification is already unblocked\n");
	return;
    }

    notification_blocked = FALSE;
    
#ifdef YOKE_HAS_SIGBLOCK

    sigsetmask(saved_sigmask);

#else /* YOKE_MAS_SIGBLOCK */

#ifdef YOKE_HAS_SIGHOLD
    
    sigrelse(NOTIFICATION_SIGNAL);

#else /* YOKE_HAS_SIGHOLD */

#error Must define either YOKE_HAS_SIGBLOCK or YOKE_HAS_SIGHOLD

#endif /* YOKE_HAS_SIGHOLD */
#endif /* YOKE_MAS_SIGBLOCK */

} /* yoke_unblock_notification() */


/*
 * notify_handler()
 *
 * This is the signal handler for NOTIFICATION_SIGNAL. If there
 * is a notification function registered, invoke it. Otherwise
 * die with a fatal error.
 */
static YOKE_SIGNAL_RETURN_T notify_handler(int s)
{
    if (notification_blocked)
	yoke_fatal("notify_handler(): Internal error: invoked when notification was blocked\n");

    if (_y_pausing_for_fatal)
    {
        yoke_printf("Received notification signal after a fatal error; still pausing\n");
        yoke_pause();
    }
    else if (notification_func != (void (*)()) NULL)
	(*notification_func)();
    else
	yoke_fatal("Received notification signal but no handler is installed.\n");
}




/*
 * yoke_kill()
 *
 * Kill the named process.
 */
void yoke_kill(int pid)
{
    kill(pid, SIGTERM);
}


/*
 * yoke_usleep()
 *
 * Sleep for usleep microseconds.
 */
void yoke_usleep(long usec)
{
    struct timeval timeout;
    
    timeout.tv_sec = usec/1000000;
    timeout.tv_usec = usec%1000000;
    
    select(0, NULL, NULL, NULL, &timeout);
} /* yoke_usleep() */
     

/*
 * yoke_wallclock()
 *
 * Return the current wallclock time in seconds.
 */
double yoke_wallclock()
{
    struct timeval tv;

    gettimeofday(&tv, 0);
    return (double) tv.tv_sec + (double) tv.tv_usec / 1000000.0;
}

/*
 * yoke_pause()
 */
void yoke_pause()
{
    yoke_printf("Process %d pausing...\n", getpid());
    pause();
} /* yoke_pause() */


#ifdef NeXT
static void malloc_error_func(int err)
{
    static char *error_codes[] =
    {
	"vm_allocate failed",
	"vm_deallocate failed",
	"vm_copy failed",
	"attempt to free or realloc freed space",
	"corrupt heap",
	"attempt to free or realloc non-allocated space",
    };

    yoke_fatal("Malloc error: %s (%d)\n",
	       error_codes[err], err);
}
#endif

char *_y_md_get_username()
{
    struct passwd *pw;
    char *local_username;
    
    pw = getpwuid(geteuid());

    if (pw == (struct passwd *) NULL)
    {
	extern char *getlogin();

	local_username = getlogin();

    }
    else
    {
	local_username = pw->pw_name;
    }
    if (local_username == NULL)
    {
	if ((local_username = getenv("USER")) == (char *) NULL)
	    local_username = "nobody";
    }
    
    return local_username;
}
