/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

/*
 * yoke.h
 *
 * This header contains the exported interface of Yoke.
 */

#include "util.h"

#if !defined(symmetry) || !defined(_TYPES_)
/* Only do this typedef if sys/types.h has not been included */
typedef int		bool_t;
#endif

#ifndef TRUE
#define	TRUE		1
#endif
#ifndef FALSE
#define	FALSE		0
#endif

#ifndef NULL
#define NULL ((void *) 0)
#endif

#ifdef YOKE_USES_GNU_MALLOC

#define malloc gnu_malloc
#define calloc gnu_calloc
#define realloc gnu_realloc
#define free gnu_free
#define memalign gnu_memalign
#define valloc gnu_valloc

#endif /* YOKE_USES_GNU_MALLOC */

#if defined(DEBUG) && !defined(__INSIGHT__)

#define YokeAssert(condition) \
    if (!(condition)) \
    { \
        yoke_fatal("Assertion " #condition " failed in file %s at line %d\n", \
		   __FILE__, __LINE__); \
    }
#define YokeAssert2(condition, print_args) \
    if (!(condition)) \
    { \
        yoke_fatal("Assertion " #condition " failed in file %s at line %d: %s", \
		   __FILE__, __LINE__, yoke_assert_sprintf print_args); \
    }
#else /* DEBUG */
#define YokeAssert(condition)
#define YokeAssert2(condition, print_args)
#endif /* DEBUG */

#ifdef MALLOC_DEBUG

#define YokeMalloc(func, var, type, size) \
    { if (((var) = (type) yoke_debug_malloc (size, __FILE__, __LINE__)) == (type) NULL) \
	yoke_fatal_malloc_6(#func, size, #type, #var, __FILE__, __LINE__); }

#define YokeFree(ptr) yoke_debug_free(ptr, __FILE__, __LINE__)

#else /* MALLOC_DEBUG */

#define YokeMalloc(func, var, type, size) \
    { if (((var) = (type) malloc (size)) == (type) NULL) \
	yoke_fatal_malloc_6(#func, size, #type, #var, __FILE__, __LINE__); }

#define YokeFree(ptr) free(ptr)

#endif /* MALLOC_DEBUG */

/*
 * Some forward declarations...
 */
typedef struct _conn_t *conn_t;
typedef union _yoke_buffer_t *yoke_buffer_t;
typedef struct _yoke_proto_t *yoke_proto_t;


/*
 * This is a function indirection table used in conn_t's
 */
typedef struct _yoke_proto_funcs_t
{
    void		(*establish)(conn_t conn);	/* This can be NULL */
    void		(*release)(conn_t conn);	/* This can be NULL */

    yoke_buffer_t	(*recv_transient)(conn_t *conn, int *type,
					  bool_t poll);
    void		(*release_transient)(conn_t conn);/* This can be NULL*/

    yoke_buffer_t	(*recv)(conn_t conn, int *type);
    void		(*destroy)(yoke_proto_t proto);
    void		(*destroy_quietly)(yoke_proto_t proto);

    yoke_proto_t	(*reinit)();
    void		(*shutdown)();
    void		(*fatal)();
    int			(*remote_start_args)(char *buf, int size);
    void		(*print)(yoke_proto_t proto);

    yoke_buffer_t	(*allocate_buffer)(conn_t dest, int size, int n_items);
#ifdef MALLOC_DEBUG
    yoke_buffer_t	(*allocate_buffer_debug)(conn_t dest, int size,
						 int n_items, char *file,
						 int line);
#endif
    int			(*sizeof_float)(conn_t conn, int count);
    int			(*sizeof_double)(conn_t conn, int count);
#ifdef YOKE_HAS_LONG_DOUBLE
    int			(*sizeof_long_double)(conn_t conn, int count);
#endif
    int			(*sizeof_short)(conn_t conn, int count);
    int			(*sizeof_u_short)(conn_t conn, int count);
    int			(*sizeof_int)(conn_t conn, int count);
    int			(*sizeof_u_int)(conn_t conn, int count);
    int			(*sizeof_long)(conn_t conn, int count);
    int			(*sizeof_u_long)(conn_t conn, int count);
    int			(*sizeof_char)(conn_t conn, int count);
    int			(*sizeof_u_char)(conn_t conn, int count);
    int			(*sizeof_address)(conn_t conn, int count);

    int  (*sizeof_proto)(conn_t dest, conn_t conn, yoke_proto_t proto);
    void (*put_proto)(yoke_buffer_t buf, conn_t conn, yoke_proto_t proto);
    void (*get_proto)(yoke_buffer_t buf, conn_t conn, yoke_proto_t *proto);
} *yoke_proto_funcs_t;


/*
 * This is a function indirection table used in yoke_buffer_t's
 */
typedef struct _yoke_proto_buffer_funcs_t
{
    void		(*send)(conn_t conn, yoke_buffer_t buf, int type);
    void		(*free_buffer)(yoke_buffer_t buf);
#ifdef MALLOC_DEBUG
    void		(*free_buffer_debug)(yoke_buffer_t buf,
					     char *file, int line);
#endif
    yoke_buffer_t	(*check_buffer_size)(yoke_buffer_t yoke_buf,
					     int slack, int increment);

    void (*put_float)(yoke_buffer_t buffer, float *data, int count);
    void (*put_double)(yoke_buffer_t buffer, double *data, int count);
#ifdef YOKE_HAS_LONG_DOUBLE
    void (*put_long_double)(yoke_buffer_t buffer, long double *data, int count);
#endif
    void (*put_short)(yoke_buffer_t buffer, short *data, int count);
    void (*put_u_short)(yoke_buffer_t buffer, unsigned short *data, int count);
    void (*put_int)(yoke_buffer_t buffer, int *data, int count);
    void (*put_u_int)(yoke_buffer_t buffer, unsigned int *data, int count);
    void (*put_long)(yoke_buffer_t buffer, long *data, int count);
    void (*put_u_long)(yoke_buffer_t buffer, unsigned long *data, int count);
    void (*put_char)(yoke_buffer_t buffer, char *data, int count);
    void (*put_u_char)(yoke_buffer_t buffer, unsigned char *data, int count);
    void (*put_address)(yoke_buffer_t buffer, void **data, int count);

    void (*get_float)(yoke_buffer_t buffer, float *dest, int count);
    void (*get_double)(yoke_buffer_t buffer, double *dest, int count);
#ifdef YOKE_HAS_LONG_DOUBLE
    void (*get_long_double)(yoke_buffer_t buffer, long double *dest, int count);
#endif
    void (*get_short)(yoke_buffer_t buffer, short *dest, int count);
    void (*get_u_short)(yoke_buffer_t buffer, unsigned short *dest, int count);
    void (*get_int)(yoke_buffer_t buffer, int *dest, int count);
    void (*get_u_int)(yoke_buffer_t buffer, unsigned int *dest, int count);
    void (*get_long)(yoke_buffer_t buffer, long *dest, int count);
    void (*get_u_long)(yoke_buffer_t buffer, unsigned long *dest, int count);
    void (*get_char)(yoke_buffer_t buffer, char *dest, int count);
    void (*get_u_char)(yoke_buffer_t buffer, unsigned char *dest, int count);
    void (*get_address)(yoke_buffer_t buffer, void **dest, int count);
} *yoke_proto_buffer_funcs_t;


/*
 * yoke_basic_buffer_t is the only part of a yoke_buffer_t that
 * needs to be exported to the user.  It is the part that is common
 * to all of the protocols that could coexist inside the yoke_buffer_t
 * union.
 *
 * This part of the yoke_buffer_t needs to live here so that
 * macros below can work.
 *
 * All protocol buffer_t's (i.e., tcp_buffer_t) should start with
 * the same fields as in yoke_basic_buffer_t.  The "union _yoke_buffer_t"
 * is defined in yoke_internal.h.
 */
typedef struct _yoke_basic_buffer_t
{
#ifdef DEBUG
    int magic;
#endif
    yoke_proto_buffer_funcs_t funcs;
} *yoke_basic_buffer_t;



/*
 * yoke_basic_proto_t is the only part of a yoke_proto_t (which is
 * part of a conn_t) that needs to be exported to the user.  It
 * just really just a stripped down version of yoke_proto_t.
 *
 * This part of the yoke_proto_t needs to live here so that
 * macros below can work.
 *
 * yoke_proto_t must start with a "yoke_proto_funct_t funcs" as
 * its first element.  This is a bit of a hack -- yoke_proto_t should
 * probably be a union.  But since its is only used in the macros
 * below for typecasting, it was deemed no worth the effort pushing
 * the union through all of yoke.  yoke_proto_t is defined
 * in yoke_internal.h.
 */
typedef struct _yoke_basic_proto_t
{
    yoke_proto_funcs_t funcs;
} *yoke_basic_proto_t;

struct _conn_t
{
    int is_transient;
    int is_established;
    int process_id;
    char *hostname_id;
    yoke_proto_t current_proto;
    yoke_proto_t protos;
    struct _conn_t *next;
    struct _conn_t *prev;
};



/*********************************************************************
 * 		Exported interface declarations
 *********************************************************************/

extern void		yoke_init(char *package_designator,
				  int  (*package_args_func)(int, int),
				  int  (*remote_start_args_func)(char *, int),
				  void (*usage_message_func)(),
				  int *n_nodes, conn_t **conn_array);
extern conn_t		yoke_fork(int *child_pid);
extern void		yoke_shutdown();
extern conn_t		yoke_self();
extern void 		yoke_get_conn_id(conn_t conn,
					 char **hostname_id,
					 int *process_id);
extern bool_t		yoke_is_established(conn_t conn);
extern yoke_buffer_t	yoke_recv_transient(conn_t *sender,
					    int *type, bool_t poll);
extern void		yoke_destroy(conn_t conn);
extern void		yoke_destroy_quietly(conn_t conn);

extern void		yoke_print(conn_t conn);
extern char *		yoke_conn_as_string(conn_t conn);

extern bool_t		yoke_conn_is_local(conn_t conn);
extern bool_t		yoke_conn_equal(conn_t conn_1, conn_t conn_2);

extern void		yoke_pause();
extern void		yoke_silent_fatal();

extern void yoke_fatal_malloc_5(char *func, char *type, char *var,
				char *file, int line);
extern void yoke_fatal_malloc_6(char *func, int size, char *type, char *var,
				char *file, int line);
#ifdef __STDC__

extern void		yoke_fatal(char *msg, ...);
extern void		yoke_error(char *msg, ...);
extern void		yoke_warning(char *msg, ...);
extern void		yoke_notice(char *msg, ...);
extern void		yoke_printf(char *msg, ...);
extern void		yoke_perror(char *msg, ...);
extern void		yoke_fatal_perror(char *msg, ...);
extern char *		yoke_assert_sprintf(char *msg, ...);

#else /* __STDC__ */

extern void		yoke_fatal();
extern void		yoke_error();
extern void		yoke_warning();
extern void		yoke_notice();
extern void		yoke_printf();
extern void		yoke_perror();
extern void		yoke_fatal_perror();
extern char *		yoke_assert_sprintf();

#endif /* __STDC__ */

/*
 * Functions that the md module defines
 */
extern void		yoke_register_notification_func(void (*func)());
extern void		yoke_register_child_death_func(void (*func)(int pid, int sig, int retcode));
extern void		yoke_notify(conn_t conn);
extern void		yoke_unblock_notification();
extern void		yoke_block_notification();
extern void 		yoke_kill(int pid);
extern void 		yoke_usleep(long usecs);
extern double		yoke_wallclock();

#ifdef MALLOC_DEBUG

extern void *		yoke_debug_malloc(int size, char *file, int line);
extern void		yoke_debug_free(void *ptr, char *file, int line);
extern bool_t		yoke_malloc_debug_level();
extern void		yoke_malloc_check(char *file, int line);
extern void		yoke_show_freed_blocks();
extern void		yoke_show_malloc_stats();

#endif


/*********************************************************************
 * 		Exported interface routines that are macros
 *********************************************************************/
/*
extern void yoke_establish(conn_t conn);
*/
#define yoke_establish(conn) \
{ \
    if (((yoke_basic_proto_t)((conn)->current_proto))->funcs->establish != NULL) \
	(((yoke_basic_proto_t)((conn)->current_proto))->funcs->establish)(conn); \
    (conn)->is_established = TRUE; \
}

/*
extern bool_t yoke_is_established(conn_t conn)
*/
#define yoke_is_established(conn) \
    ((conn)->is_established)

/*
extern void yoke_release(conn_t conn);
*/
#define yoke_release(conn) \
{ \
    if (((yoke_basic_proto_t)((conn)->current_proto))->funcs->release != NULL) \
	(((yoke_basic_proto_t)((conn)->current_proto))->funcs->release)(conn); \
    (conn)->is_established = FALSE; \
}

/*
extern void yoke_release_transient(conn_t conn);
*/
#define yoke_release_transient(conn) \
{ \
    if (((yoke_basic_proto_t)((conn)->current_proto))->funcs->release_transient != NULL) \
	(((yoke_basic_proto_t)((conn)->current_proto))->funcs->release_transient)(conn); \
    (conn)->is_established = FALSE; \
}

/*
extern void		yoke_send(conn_t conn, yoke_buffer_t buf, int type);
*/
#define yoke_send(conn, buf, type) \
    (((yoke_basic_buffer_t)(buf))->funcs->send)((conn), (buf), (type))

/*
extern yoke_buffer_t	yoke_recv(conn_t conn, int *type);
*/
#define yoke_recv(conn, type) \
    (((yoke_basic_proto_t)((conn)->current_proto))->funcs->recv)(conn, type)


/*********************************************************************
 * 		Buffer management routines
 *********************************************************************/

/*
extern yoke_buffer_t yoke_allocate_buffer(conn_t dest, int size, int n_items);
extern void yoke_free_buffer(yoke_buffer_t buf);
extern yoke_buffer_t yoke_check_buffer_size(yoke_buffer_t buffer,
					    int slack, int increment);
*/

#ifdef MALLOC_DEBUG

#define yoke_allocate_buffer(dest, size, n_items) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->allocate_buffer_debug)(dest, size, n_items, __FILE__, __LINE__)

#define yoke_free_buffer(buf) \
    (((yoke_basic_buffer_t)(buf))->funcs->free_buffer_debug)(buf, __FILE__, __LINE__)

#else /* MALLOC_DEBUG */

#define yoke_allocate_buffer(dest, size, n_items) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->allocate_buffer)(dest, size, n_items)

#define yoke_free_buffer(buf) \
    (((yoke_basic_buffer_t)(buf))->funcs->free_buffer)(buf)

#endif /* MALLOC_DEBUG */

#define yoke_check_buffer_size(buf, slack, increment) \
    (((yoke_basic_buffer_t)(buf))->funcs->check_buffer_size)(buf, slack, increment)


/*
extern void yoke_put_float(yoke_buffer_t buffer, float *data, int count);
extern void yoke_put_double(yoke_buffer_t buffer, double *data, int count);
extern void yoke_put_short(yoke_buffer_t buffer, short *data, int count);
extern void yoke_put_u_short(yoke_buffer_t buffer, unsigned short *data,
			     int count);
extern void yoke_put_int(yoke_buffer_t buffer, int *data, int count);
extern void yoke_put_u_int(yoke_buffer_t buffer, unsigned int *data,
			   int count);
extern void yoke_put_long(yoke_buffer_t buffer, long *data, int count);
extern void yoke_put_u_long(yoke_buffer_t buffer, unsigned long *data,
			    int count);
extern void yoke_put_char(yoke_buffer_t buffer, char *data, int count);
extern void yoke_put_u_char(yoke_buffer_t buffer, unsigned char *data,
			    int count);
extern void yoke_put_address(yoke_buffer_t buffer, void **addr, int count);
*/
#define yoke_put_float(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_float)(buf, data, count)
#define yoke_put_double(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_double)(buf, data, count)
#ifdef YOKE_HAS_LONG_DOUBLE
#define yoke_put_long_double(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_long_double)(buf, data, count)
#endif
#define yoke_put_short(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_short)(buf, data, count)
#define yoke_put_u_short(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_u_short)(buf, data, count)
#define yoke_put_int(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_int)(buf, data, count)
#define yoke_put_u_int(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_u_int)(buf, data, count)
#define yoke_put_long(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_long)(buf, data, count)
#define yoke_put_u_long(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_u_long)(buf, data, count)
#define yoke_put_char(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_char)(buf, data, count)
#define yoke_put_u_char(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_u_char)(buf, data, count)
#define yoke_put_address(buf, data, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->put_address)(buf, data, count)



/*
extern void yoke_get_float(yoke_buffer_t buffer, float *dest, int count);
extern void yoke_get_double(yoke_buffer_t buffer, double *dest, int count);
extern void yoke_get_short(yoke_buffer_t buffer, short *dest, int count);
extern void yoke_get_u_short(yoke_buffer_t buffer, unsigned short *dest,
			     int count);
extern void yoke_get_int(yoke_buffer_t buffer, int *dest, int count);
extern void yoke_get_u_int(yoke_buffer_t buffer, unsigned int *dest,
			   int count);
extern void yoke_get_long(yoke_buffer_t buffer, long *dest, int count);
extern void yoke_get_u_long(yoke_buffer_t buffer, unsigned long *dest,
			    int count);
extern void yoke_get_char(yoke_buffer_t buffer, char *dest, int count);
extern void yoke_get_u_char(yoke_buffer_t buffer, unsigned char *dest,
			    int count);
extern void yoke_get_address(yoke_buffer_t buffer, void **dest, int count);
*/
#define yoke_get_float(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_float)(buf, dest, count)
#define yoke_get_double(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_double)(buf, dest, count)
#ifdef YOKE_HAS_LONG_DOUBLE
#define yoke_get_long_double(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_long_double)(buf, dest, count)
#endif
#define yoke_get_short(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_short)(buf, dest, count)
#define yoke_get_u_short(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_u_short)(buf, dest, count)
#define yoke_get_int(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_int)(buf, dest, count)
#define yoke_get_u_int(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_u_int)(buf, dest, count)
#define yoke_get_long(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_long)(buf, dest, count)
#define yoke_get_u_long(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_u_long)(buf, dest, count)
#define yoke_get_char(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_char)(buf, dest, count)
#define yoke_get_u_char(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_u_char)(buf, dest, count)
#define yoke_get_address(buf, dest, count) \
    (((yoke_basic_buffer_t)(buf))->funcs->get_address)(buf, dest, count)


/*
extern int	yoke_sizeof_float(conn_t dest, int count);
extern int	yoke_sizeof_double(conn_t dest, int count);
extern int	yoke_sizeof_short(conn_t dest, int count);
extern int	yoke_sizeof_u_short(conn_t dest, int count);
extern int	yoke_sizeof_int(conn_t dest, int count);
extern int	yoke_sizeof_u_int(conn_t dest, int count);
extern int	yoke_sizeof_long(conn_t dest, int count);
extern int	yoke_sizeof_u_long(conn_t dest, int count);
extern int	yoke_sizeof_char(conn_t dest, int count);
extern int	yoke_sizeof_u_char(conn_t dest, int count);
extern int	yoke_sizeof_address(conn_t dest, int count);
*/
#define yoke_sizeof_float(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_float)(dest, count)
#define yoke_sizeof_double(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_double)(dest, count)
#ifdef YOKE_HAS_LONG_DOUBLE
#define yoke_sizeof_long_double(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_long_double)(dest, count)
#endif
#define yoke_sizeof_short(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_short)(dest, count)
#define yoke_sizeof_u_short(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_u_short)(dest, count)
#define yoke_sizeof_int(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_int)(dest, count)
#define yoke_sizeof_u_int(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_u_int)(dest, count)
#define yoke_sizeof_long(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_long)(dest, count)
#define yoke_sizeof_u_long(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_u_long)(dest, count)
#define yoke_sizeof_char(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_char)(dest, count)
#define yoke_sizeof_u_char(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_u_char)(dest, count)
#define yoke_sizeof_address(dest, count) \
    (((yoke_basic_proto_t)((dest)->current_proto))->funcs->sizeof_address)(dest, count)


extern int	yoke_sizeof_conn(conn_t dest, conn_t conn);
extern void	yoke_put_conn(yoke_buffer_t buffer, conn_t conn);
extern void	yoke_get_conn(yoke_buffer_t buffer, conn_t *new_conn);

